import re
import os
from Bio.PDB import PDBParser, Selection
from Bio import SeqIO, pairwise2
import types
from collections import OrderedDict

dictaa_3letters = {
    "P": "PRO",
    "S": "SER",
    "G": "GLY",
    "L": "LEU",
    "T": "THR",
    "V": "VAL",
    "I": "ILE",
    "M": "MET",
    "C": "CYS",
    "F": "PHE",
    "Y": "TYR",
    "W": "TRP",
    "A": "ALA",
    "K": "LYS",
    "R": "ARG",
    "D": "ASP",
    "E": "GLU",
    "N": "ASN",
    "Q": "GLN",
    "H": "HIS",
}

dictaa_1letter = {
    "VAL": "V",
    "ILE": "I",
    "LEU": "L",
    "GLU": "E",
    "GLN": "Q",
    "ASP": "D",
    "ASN": "N",
    "HIS": "H",
    "TRP": "W",
    "PHE": "F",
    "TYR": "Y",
    "ARG": "R",
    "LYS": "K",
    "SER": "S",
    "THR": "T",
    "MET": "M",
    "ALA": "A",
    "GLY": "G",
    "PRO": "P",
    "CYS": "C",
}

dssp_sse_desc = {
    "H": "helix",
    "E": "strand",
    "T": "turn",
    "B": "beta bridge",
    "G": "helix-3",
    "I": "helix-5",
    "S": "bend",
}
# the mutation results in loss of polar atoms in the side-chain
# and interactions involving them
list_residues_polar = ["T", "S", "H", "N", "Q", "D", "E", "K", "R", "Y", "W"]
# the mutation results in loss of aromatic side-chain and interactions
# involving the aromatic ring
list_aromatic_residues = ["F", "Y", "W"]
# the mutation results in hydrophobic to polar side-chain properties
list_residues_hydrophobic = ["G", "A", "L", "V", "I", "P", "M", "F", "C"]
list_residues_backbone_unique = ["P", "G"]
# In the selected structure, the polar side-chain atom is involved in interactions
dict_atoms_acceptor = {
    "T": ["OG1"],
    "S": ["OG"],
    "Y": ["OH"],
    "N": ["OD1"],
    "Q": ["OE1"],
    "D": ["OD1", "OD2"],
    "E": ["OE1", "OE2"],
    "H": ["ND1", "NE2"],
}
dict_atoms_donor = {
    "T": ["OG1"],
    "S": ["OG"],
    "Y": ["OH"],
    "N": ["ND2"],
    "Q": ["NE2"],
    "H": ["ND1", "NE2"],
    "R": ["NE", "NH1", "NH2"],
    "K": ["NZ"],
    "W": ["NE1"],
}
list_branched_sidechains = ["S", "T", "I", "V", "L"]
list_flexible_sidechains = ["M", "K"]
dict_sidechain_accessible_surface = {
    "A": 67,
    "R": 196,
    "N": 113,
    "D": 106,
    "C": 104,
    "Q": 144,
    "E": 138,
    "G": 1,
    "H": 151,
    "I": 140,
    "L": 137,
    "K": 167,
    "M": 160,
    "F": 175,
    "P": 105,
    "S": 80,
    "T": 102,
    "W": 217,
    "Y": 187,
    "V": 117,
}
dict_pI_aa = {
    "ARG": 10.76,
    "LYS": 9.74,
    "HIS": 7.59,
    "GLY": 5.97,
    "ALA": 6.00,
    "VAL": 5.96,
    "LEU": 5.98,
    "ILE": 6.02,
    "PRO": 6.30,
    "MET": 5.74,
    "SER": 5.68,
    "THR": 5.60,
    "ASN": 5.41,
    "GLN": 5.65,
    "ASP": 2.77,
    "GLU": 3.22,
    "PHE": 5.48,
    "TYR": 5.66,
    "TRP": 5.89,
    "CYS": 5.07,
}
# Rose et al. 1985
dict_maxacc = {
    "A": 118.1,
    "R": 256.0,
    "N": 165.5,
    "D": 158.7,
    "C": 146.1,
    "E": 186.2,
    "Q": 193.2,
    "G": 88.1,
    "H": 202.5,
    "I": 181.0,
    "L": 193.1,
    "K": 225.8,
    "M": 203.4,
    "F": 222.8,
    "P": 146.8,
    "S": 129.8,
    "T": 152.5,
    "W": 266.3,
    "Y": 236.8,
    "V": 164.5,
}


class pdb2seq:
    """
    Functions to parse pdbfile and extract sequences
    """

    def __init__(self, pdbfile):
        self.pdbfile = pdbfile

    def get_pdbatomseq(self):
        """
        get seq based on atom records
        """
        dict_seq = {}
        dict_resdet = {}
        pdbname = os.path.split(self.pdbfile)[1].split(".")[0]
        pdbobj = PDBParser(QUIET=True)
        structure_instance = pdbobj.get_structure(pdbname, self.pdbfile)
        for model in structure_instance:
            for chain in model:
                chid = chain.get_id()
                if chid in ["", " "]:
                    chid_fix = " "
                else:
                    chid_fix = chid
                residues = Selection.unfold_entities(chain, "R")
                seq = ""
                list_resdet = []
                for res in residues:
                    hetatm_flag, resnum, icode = res.get_id()
                    resname = res.get_resname()
                    if resname not in dictaa_1letter:
                        continue
                    seq += dictaa_1letter[resname]
                    list_resdet.append(str(resnum) + " " + dictaa_1letter[resname])

                dict_seq[pdbname + "_" + chid_fix] = seq
                dict_resdet[pdbname + "_" + chid_fix] = list_resdet[:]
        return dict_seq, dict_resdet

    def get_pdbseqres(self, outfile=None, chain_id=None, flagout=False, warnings=False):
        """
        get seq based on seqres if available
        return '' otherwise
        """
        dict_seq = {}
        pdbname = os.path.split(self.pdbfile)[1].split(".")[0]
        list_records = SeqIO.parse(self.pdbfile, "pdb-seqres")
        if list_records is None or not isinstance(list_records, types.GeneratorType):
            print("Warning: seqres record not found")
            return dict_seq
        for record in list_records:
            # chain ID?
            if chain_id is not None:
                if record.annotations["chain"] != chain_id:
                    continue
            else:
                chain_id = record.annotations["chain"]
            if flagout:
                outfile = os.path.splitext(self.pdbfile)[0] + "_" + chain_id + ".fasta"
                ou = open(os.path.abspath(outfile), "w")
                ou.write(">" + pdbname + "_" + chain_id + "\n")
            str_seq = ""
            if re.search("X", str(record.seq)) is not None:
                for s in str(record.seq):
                    if s != "X":
                        str_seq += s
            else:
                str_seq = str(record.seq)
            if warnings:
                if len(str_seq) > 800:
                    print("Sequence too long for Jpred: {}".format(chain_id))
                    continue
                elif len(str_seq) < 30:
                    print("Sequence too short for Jpred: {}".format(chain_id))
                    continue
            if flagout:
                ou.write(str_seq + "\n")
                ou.close()
            dict_seq[pdbname + "_" + record.annotations["chain"]] = str_seq
        return dict_seq

    def match_atom_seqres(self, dict_atomseq, dict_seqres):
        dict_atomseq_aln = {}
        dict_seqres_aln = {}
        for pdbch in dict_atomseq:
            if pdbch not in dict_seqres:
                continue
            aligned_seq1, aligned_seq2 = alignseq(
                dict_atomseq[pdbch], dict_seqres[pdbch]
            )
            # skip non similar sequences
            if float(aligned_seq1.count("-")) / len(aligned_seq1) < 0.5:
                continue
            if float(aligned_seq2.count("-")) / len(aligned_seq2) < 0.5:
                continue
            dict_atomseq_aln[pdbch] = aligned_seq1[:]
            dict_seqres_aln[pdbch] = aligned_seq2[:]

        return dict_atomseq_aln, dict_seqres_aln


def alignseq(seq1, seq2):
    if isinstance(seq1, list):
        seq1_str = "".join(seq1)
    elif isinstance(seq1, str):
        seq1_str = seq1
    else:
        print("Could not align sequences:")
        print(seq1)
        print(seq2)
        return None, None
    if isinstance(seq2, list):
        seq2_str = "".join(seq2)
    elif isinstance(seq2, str):
        seq2_str = seq2
    else:
        print("Could not align sequences:")
        print(seq1)
        print(seq2)
        return None, None
    aligned = pairwise2.align.globalms(seq1_str, seq2_str, 1, -1, -0.5, -0.1)[0]
    aligned_seq1, aligned_seq2 = list(aligned[0]), list(aligned[1])
    return aligned_seq1, aligned_seq2


def match_lists_add(list1, list2):
    if isinstance(list1, str):
        list1 = [e for e in list1]
    if isinstance(list2, str):
        list2 = [e for e in list2]
    # add gaps to list1 based on list2
    indices = [i for i, x in enumerate(list2) if x == "-"]
    ct = 0
    for i in indices:
        list1.insert(i, "-")
        ct += 1
    return list1


def match_lists_del(list1, list2):
    if isinstance(list1, str):
        list1 = [e for e in list1]
    if isinstance(list2, str):
        list2 = [e for e in list2]
    # remove indices in list1 corresponding to gaps in list2
    indices = [i for i, x in enumerate(list2) if x == "-"]
    ct = 0
    for i in indices:
        list1.pop(i - ct)
        ct += 1
    return list1


def parse_dssp_file(dsspfile, modelid):
    out_fh = open(dsspfile, "r")
    residue_scores_start = False
    dict_sses = {}
    for line in out_fh:
        if residue_scores_start:
            if modelid not in dict_sses:
                dict_sses[modelid] = {}
            parse_dssp_line(line[:-1], modelid, dict_sses)
        if "#" in line and "RESIDUE" in line:
            residue_scores_start = True
    return dict_sses


def parse_dssp_line(line, modelid, dict_sses):
    line_split = line.split()
    # residue number
    resnum = line_split[1]
    # residue name
    try:
        resname = dictaa_3letters[line_split[3]]
    except KeyError:
        resname = line_split[3]
    # chain ID
    chainID = line_split[2]
    # secondary structure
    dssp_sse = line[16]
    # secondary structure description
    try:
        sse_desc = dssp_sse_desc[dssp_sse]
    except KeyError:
        sse_desc = " "
    # calculate relative solvant accessibility
    try:
        dssp_acc = float(line[35:38])
        residue_1l = line_split[3]
        if len(residue_1l) == 1 and residue_1l.islower():
            residue_1l = "C"  # disulfide bridge
        dssp_acc = dssp_acc / dict_maxacc[residue_1l]
    except KeyError:
        # print sse_desc,dssp_sse,dssp_acc,resname,line_split[3]
        dssp_acc = -1.0
    # Ca coordinates
    try:
        x, y, z = ["{:.4f}".format(float(coord)) for coord in line_split[-5:-2]]
    except IndexError:
        x = y = z = ""
    # save details
    try:
        dict_sses[modelid][chainID][resnum] = [
            sse_desc,
            dssp_sse,
            dssp_acc,
            resname,
            x,
            y,
            z,
        ]
    except KeyError:
        dict_sses[modelid][chainID] = OrderedDict()
        dict_sses[modelid][chainID][resnum] = [
            sse_desc,
            dssp_sse,
            dssp_acc,
            resname,
            x,
            y,
            z,
        ]


def rgb(minimum, maximum, value):
    minimum, maximum = float(minimum), float(maximum)
    ratio = 2 * (value - minimum) / (maximum - minimum)
    b = int(max(0, 255 * (1 - ratio)))
    r = int(max(0, 255 * (ratio - 1)))
    g = 255 - b - r
    return r, g, b



#
#     Copyright (C) 2021 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.
#

import unittest
import os
import shutil
import tempfile
import math
from ccpem_utils_tests import test_data
import subprocess
from ccpem_pyutils.scripts import get_map_parameters


class MapParseTests(unittest.TestCase):
    def setUp(self):
        """
        Setup test data and output directories.
        """
        self.test_data = os.path.dirname(test_data.__file__)
        self.test_dir = tempfile.mkdtemp(prefix="map_parse")
        # Change to test directory
        self._orig_dir = os.getcwd()
        os.chdir(self.test_dir)

    def tearDown(self):
        os.chdir(self._orig_dir)
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_run_subprocess_get_map_parameters(self):
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        subprocess.call(
            [
                "python3 "
                + os.path.realpath(get_map_parameters.__file__)
                + " -m "
                + map_input
                + " -odir "
                + self.test_dir,
            ],
            shell=True,
        )
        assert os.path.isfile(
            os.path.join(self.test_dir, "emd_3488_map_parameters.json")
        )
        assert math.isclose(
            os.stat(
                os.path.join(self.test_dir, "emd_3488_map_parameters.json")
            ).st_size,
            275,
            rel_tol=0.05,
        )



import argparse
import os
import json
from ccpem_pyutils.model.gemmi_model_utils import (
    GemmiModelUtils,
    set_bfactor_attributes,
)


def parse_args():
    parser = argparse.ArgumentParser(description="CCP-EM analyse B-factors")
    parser.add_argument(
        "-p",
        "--model",
        required=True,
        help="Input atomic model (.pdb, .cif)",
    )
    parser.add_argument(
        "-odir",
        "--odir",
        required=False,
        default=None,
        help="Output directory",
    )
    parser.add_argument(
        "-mid",
        "--mid",
        required=False,
        default=None,
        help="Model ID to prefix output file names",
    )

    return parser.parse_args()


def get_residue_bfactors(
    gemmiutils: GemmiModelUtils, modelid: str = None, outdir: str = None
):

    dict_bfact_dev = gemmiutils.get_avgbfact_deviation()
    dict_attr = {}
    max_bfact_dev = 0.0
    for model in dict_bfact_dev:
        dict_attr = {}
        for chain in dict_bfact_dev[model]:
            for res_id in dict_bfact_dev[model][chain]:
                residue_id = "_".join([model, chain, res_id])
                dict_attr[residue_id] = dict_bfact_dev[model][chain][res_id][1]
                max_bfact_dev = max(
                    dict_bfact_dev[model][chain][res_id][1], max_bfact_dev
                )
        break
    # maximum deviation
    dict_bfact_dev["max_dev"] = max_bfact_dev
    if outdir:
        out_json = os.path.join(outdir, modelid + "_residue_bfactors.json")
    else:
        out_json = modelid + "_residue_bfactors.json"
    with open(out_json, "w") as j:
        json.dump(dict_bfact_dev, j)
    return dict_attr


def get_residue_coordinates(
    gemmiutils: GemmiModelUtils, modelid: str = None, outdir: str = None
):

    dict_ca_coord = gemmiutils.get_coordinates(atom_selection="one_per_residue")
    if outdir:
        out_json = os.path.join(outdir, modelid + "_residue_coordinates.json")
    else:
        out_json = modelid + "_residue_coordinates.json"
    with open(out_json, "w") as j:
        json.dump(dict_ca_coord, j)


def get_resolution(
    gemmiutils: GemmiModelUtils, modelid: str = None, outdir: str = None
):

    dict_resolution = {"resolution": gemmiutils.resolution}
    if outdir:
        out_json = os.path.join(outdir, modelid + "_model_info.json")
    else:
        out_json = modelid + "_model_info.json"
    with open(out_json, "w") as j:
        json.dump(dict_resolution, j)


def get_residue_names(
    gemmiutils: GemmiModelUtils, modelid: str = None, outdir: str = None
):
    gemmiutils.set_residue_types()
    dict_residue_names = gemmiutils.dict_resnames
    if outdir:
        out_json = os.path.join(outdir, modelid + "_residue_names.json")
    else:
        out_json = modelid + "_residue_names.json"
    with open(out_json, "w") as j:
        json.dump(dict_residue_names, j)


def main():
    args = parse_args()
    # read input
    modelfile = args.model
    if args.mid:
        modelid = args.mid
    else:
        modelid = os.path.splitext(os.path.basename(modelfile))[0]
    gemmiutils = GemmiModelUtils(modelfile)
    # get b-factors
    dict_attr = get_residue_bfactors(gemmiutils, modelid=modelid, outdir=args.odir)
    set_bfactor_attributes(modelfile, dict_attr, attr_name="bfactdev")
    # get coordinates
    get_residue_coordinates(gemmiutils, modelid=modelid, outdir=args.odir)
    # get resolution
    get_resolution(gemmiutils, modelid=modelid, outdir=args.odir)
    # get residue names
    get_residue_names(gemmiutils, modelid=modelid, outdir=args.odir)


if __name__ == "__main__":
    main()



#
#     Copyright (C) 2021 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.
#

import unittest
import os
import shutil
import tempfile
import math
from ccpem_utils_tests import test_data
import subprocess
from ccpem_pyutils.scripts import get_map_parameters
from ccpem_pyutils.map.parse_mrcmapobj import MapObjHandle
from ccpem_pyutils.model.coord_grid import mapGridPositions
from ccpem_pyutils.other.calc import get_ccc
import mrcfile


class MapParseTests(unittest.TestCase):
    def setUp(self):
        """
        Setup test data and output directories.
        """
        self.test_data = os.path.dirname(test_data.__file__)
        self.test_dir = tempfile.mkdtemp(prefix="map_parse")
        # Change to test directory
        self._orig_dir = os.getcwd()
        os.chdir(self.test_dir)

    def tearDown(self):
        os.chdir(self._orig_dir)
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_map_parse(self):
        # read
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        with mrcfile.open(map_input, mode="r", permissive=True) as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
        # process
        wrapped_mapobj.shift_origin((1.0, 1.0, 1.0))
        with mrcfile.new("emd_3488_shifted.mrc", overwrite=True) as mrc:
            wrapped_mapobj.update_newmap_data_header(mrc)
        wrapped_mapobj.close()
        # write
        with mrcfile.open("emd_3488_shifted.mrc", mode="r") as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
            assert wrapped_mapobj.origin == (1.0, 1.0, 1.0)

    def test_map_fix_origin(self):
        # read
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        with mrcfile.open(map_input, mode="r", permissive=True) as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
        # process
        wrapped_mapobj.shift_nstart((1, 1, 1))
        wrapped_mapobj.fix_origin()
        assert tuple([round(o, 5) for o in wrapped_mapobj.origin]) == (1.05, 1.05, 1.05)

    def test_map_crop(self):
        # read
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        with mrcfile.open(map_input, mode="r", permissive=True) as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
        # copy
        wrapped_mapobj_copy = wrapped_mapobj.copy()
        # process1
        wrapped_mapobj.crop_map_grid(new_dim=(71, 73, 58))
        # write1
        with mrcfile.new("emd_3488_cropped1.mrc", overwrite=True) as mrc:
            wrapped_mapobj.update_newmap_data_header(mrc)
        wrapped_mapobj.close()
        # check1
        with mrcfile.open("emd_3488_cropped1.mrc", mode="r") as mrc:
            wrapped_mapobj1 = MapObjHandle(mrc)
            assert wrapped_mapobj1.data.shape == (58, 73, 71)
        # process2
        wrapped_mapobj2 = wrapped_mapobj_copy.crop_map_grid(
            contour=0.125, ext=3, inplace=False
        )
        # write
        with mrcfile.new("emd_3488_cropped2.mrc", overwrite=True) as mrc:
            wrapped_mapobj2.update_newmap_data_header(mrc)
        # check
        with mrcfile.open("emd_3488_cropped2.mrc", mode="r") as mrc:
            wrapped_mapobj2 = MapObjHandle(mrc)
            assert wrapped_mapobj2.data.shape == (58, 73, 71)
        assert (
            get_ccc(
                wrapped_mapobj1.data,
                wrapped_mapobj2.data,
            )
            == 1.0
        )
        wrapped_mapobj_copy.crop_map_grid(contour=0.125, ext=3, cubic=True)
        assert wrapped_mapobj_copy.data.shape == (73, 73, 73)

    def test_map_pad(self):
        # read
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        with mrcfile.open(map_input, mode="r", permissive=True) as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
        # process
        wrapped_mapobj.pad_map_grid(ext_dim=(10, 10, 10))
        min_map = wrapped_mapobj.data.min()
        # write
        with mrcfile.new("emd_3488_padded.mrc", overwrite=True) as mrc:
            wrapped_mapobj.update_newmap_data_header(mrc)
        wrapped_mapobj.close()
        # check1
        with mrcfile.open("emd_3488_padded.mrc", mode="r") as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
            assert wrapped_mapobj.data.shape == (110, 110, 110)
        min_padded_map = wrapped_mapobj.data.min()
        assert min_map == min_padded_map  # padding filled with min

    def test_map_coordinate_search(self):
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        with mrcfile.open(map_input, mode="r", permissive=True) as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
            list_grid_points = mapGridPositions(
                wrapped_mapobj, atom_coord=(61.197, 39.327, 61.266), res_map=3.2
            )
            assert len(list_grid_points) == 27
            list_grid_points = mapGridPositions(
                wrapped_mapobj,
                atom_coord=(61.197, 39.327, 61.266),
                res_map=3.2,
                gauss=False,
            )
            assert len(list_grid_points) == 1

    def test_run_subprocess_get_map_parameters(self):
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        subprocess.call(
            [
                "python3 "
                + os.path.realpath(get_map_parameters.__file__)
                + " -m "
                + map_input
                + " -odir "
                + self.test_dir,
            ],
            shell=True,
        )
        assert os.path.isfile(
            os.path.join(self.test_dir, "emd_3488_map_parameters.json")
        )
        assert math.isclose(
            os.stat(
                os.path.join(self.test_dir, "emd_3488_map_parameters.json")
            ).st_size,
            275,
            rel_tol=0.05,
        )



from sklearn import cluster, mixture
from sklearn.preprocessing import MinMaxScaler
from sklearn.neighbors import KDTree
from typing import Union
import numpy as np


def cluster_coord_features(
    array_3Dcoord_features: Union[list, np.ndarray],
    num_clusters: int = None,
    dbscan_eps: float = 2.3,
    norm: bool = False,
) -> np.ndarray:
    """
    Cluster spatial features based on
        1) feature density (DBSCAN) or
        2) gaussian mixture model (GMM) when number of clusters is known

    Arguments
    ---------
        array_3Dcoord_features: np.ndarray
            input array (e.g. [[x1,y1,z1,feature11,feature12,...],
                                [x2,y2,z2,feature21,feature22,...]])
        num_clusters: int, optional
            number of clusters expected
        dbscan_eps: float, optional
            maximum distance within same neighborhood (not within a cluster)
            (used only when the number of clusters is not provided)
            Useful eps values for 3D protein coordinate domain level separation:
            all atom: 2.3 (norm - 0.035)
            backbone atoms: 2.8 (norm - 0.04)
            c-alpha: 3.83 [3.7 - 4.0] (norm - 0.07673)
            centre: 5.35 [5.2 - 5.45] (norm - 0.088), geometric centre.

        norm: bool, optional
            Normalize each column (feature) to a similar range
    Returns
    -------
        Array with cluster labels for each input row
    """
    if isinstance(array_3Dcoord_features, list):
        array_3Dcoord_features = np.array(array_3Dcoord_features)
    if norm:
        scaler = MinMaxScaler()
        scaler.fit(array_3Dcoord_features)
        # scale each feature in [0,1], normalising by min and max of each feature
        array_3Dcoord_features_norm = scaler.transform(array_3Dcoord_features)
        if dbscan_eps == 2.3:
            dbscan_eps = 0.035
    else:
        array_3Dcoord_features_norm = array_3Dcoord_features
    if num_clusters is not None:
        model = mixture.GaussianMixture(
            n_components=num_clusters, covariance_type="full"
        )
        model.fit(array_3Dcoord_features_norm)
        return model.predict(array_3Dcoord_features_norm)
    else:
        model = cluster.DBSCAN(eps=dbscan_eps, min_samples=2)
        model.fit_predict(array_3Dcoord_features_norm)
        return model.labels_


def generate_kdtree(
    coords: Union[list, np.ndarray], leaf_size: int = None, metric: str = "euclidean"
) -> KDTree:
    """
    Get a KDtree object from input coordinates

    Arguments
    ---------
        coords: list or np.ndarray
            input array (e.g. [[x1,y1,z1],
                                [x2,y2,z2]])
        leaf_size: int, optional
            size or each segment/leaf of the tree
            tree building is faster with larger leaf size
            querying is usually faster when leaf_size is comparable
                (say between k and 3*k) to the number of neighbors queried
        metric: string, optional
            distance metric, euclidean by default
    Returns
    -------
        sklearn KDTree object
    """
    if leaf_size:
        return KDTree(coords, leaf_size=leaf_size, metric=metric)
    else:
        return KDTree(coords, metric=metric)


def get_neighbors_kdtree(
    coords: Union[list, np.ndarray],
    kdtree: KDTree,
    distance: float = None,
    num_neighbors: int = None,
) -> np.ndarray:
    if distance:
        return kdtree.query_radius(coords, distance)
    elif num_neighbors:
        return kdtree.query(coords, num_neighbors)


def pairs_kdtree(
    coords: Union[list, np.ndarray], kdtree: KDTree, dist: float
) -> np.ndarray:

    neighbor_index = get_neighbors_kdtree(coords, kdtree, distance=dist)
    pair_indices_array = np.array([], dtype=int)
    ln = 0
    for neighbors in neighbor_index:
        for n in neighbors:
            pair_index = np.array([[ln, n]])
            if pair_indices_array.size == 0:
                pair_indices_array = pair_index
            else:
                pair_indices_array = np.append(pair_indices_array, pair_index, axis=0)
        ln += 1
    return pair_indices_array



import argparse
import mrcfile
import os
import numpy as np
import json
import pathlib
from typing import Union


def parse_args():
    parser = argparse.ArgumentParser(description="CCP-EM MRC parse")
    parser.add_argument(
        "-m",
        "--map",
        required=True,
        help="Input map (MRC)",
    )
    parser.add_argument(
        "-odir",
        "--odir",
        required=False,
        default=None,
        help="Output directory",
    )

    return parser.parse_args()


def get_mrc_header_dataparameters(
    mapfile: Union[str, pathlib.Path], outdir: Union[str, pathlib.Path, None] = None
):
    map_parameters = {}
    with mrcfile.open(mapfile, mode="r", permissive=True) as mrc:
        map_parameters["nx"] = mrc.header.nx.item()
        map_parameters["ny"] = mrc.header.ny.item()
        map_parameters["nz"] = mrc.header.nz.item()
        map_parameters["nxstart"] = mrc.header.nxstart.item()
        map_parameters["nystart"] = mrc.header.nystart.item()
        map_parameters["nzstart"] = mrc.header.nzstart.item()
        map_parameters["originx"] = str(round(mrc.header.origin.x.item(), 3))
        map_parameters["originy"] = str(round(mrc.header.origin.y.item(), 3))
        map_parameters["originz"] = str(round(mrc.header.origin.z.item(), 3))
        map_parameters["cella"] = (
            str(round(mrc.header.cella.x.item(), 3)),
            str(round(mrc.header.cella.y.item(), 3)),
            str(round(mrc.header.cella.z.item(), 3)),
        )
        map_parameters["apix"] = (
            str(round(mrc.voxel_size.item()[0], 5)),
            str(round(mrc.voxel_size.item()[1], 5)),
            str(round(mrc.voxel_size.item()[2], 5)),
        )
        map_parameters["min"] = str(round(np.amin(mrc.data), 5))
        map_parameters["max"] = str(round(np.amax(mrc.data), 5))
        map_parameters["mean"] = str(round(np.mean(mrc.data), 5))
        map_parameters["std"] = str(round(np.std(mrc.data), 5))

        map_basename = os.path.splitext(os.path.basename(mapfile))[0]
        if outdir:
            out_json = os.path.join(outdir, map_basename + "_map_parameters.json")
        else:
            out_json = map_basename + "_map_parameters.json"

        with open(out_json, "w") as j:
            json.dump(map_parameters, j)


def main():
    args = parse_args()
    get_mrc_header_dataparameters(args.map, outdir=args.odir)


if __name__ == "__main__":
    main()



#
#     Copyright (C) 2019 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.

import os
import json
from typing import OrderedDict, Union, List
import pathlib
import gemmi
import numpy as np
import copy


class GemmiStructureExt(object):
    """
    Helper class to extend gemmi Structure class
    """

    def __init__(self, parent_structure: gemmi.Structure):
        self.parent_structure = parent_structure
        for attr in dir(parent_structure):
            try:
                self.__dict__[attr] = copy.deepcopy(self.__getattribute__(attr))
            except TypeError:
                pass

    def __getattribute__(self, name: str):
        try:
            # self.parent_structure call results in recursion error here
            # use superclass method instead
            return getattr(
                super(GemmiStructureExt, self).__getattribute__("parent_structure"),
                name,
            )
        except AttributeError:
            return super(GemmiStructureExt, self).__getattribute__(name)


class GemmiModelUtils(object):
    """
    Operations on gemmi Structure object
    """

    def __init__(self, input_model: Union[pathlib.Path, str, gemmi.Structure]):

        self.structure = check_structure_inputs(
            input_model
        )  # gemmi.read_structure(modelfile)
        self.resolution = self.structure.resolution
        if os.path.isfile(input_model):
            self.file_ext = os.path.splitext(input_model)[1].lower()
            self.modelid = os.path.splitext(os.path.basename(input_model))[0]
        else:
            self.file_ext = None
            self.modelid = None

    def write_structure_as_mmcif(self, mmcif_name):
        """Write a Gemmi structure out to an mmCIF file."""
        write_structure_as_mmcif(self.structure, mmcif_name)

    def write_structure_as_pdb(self, pdb_name):
        """Write a Gemmi structure out to an PDB file."""
        write_structure_as_pdb(self.structure, pdb_name)

    def shift_coordinates(
        self,
        trans_vector: Union[list, np.ndarray, tuple],
        out_model_path: Union[pathlib.Path, str] = None,
        remove_charges: bool = False,
        inplace: bool = False,
    ):
        """
        Shift atomic coordinates based on a translation vector (x,y,z)
        """
        if inplace:
            structure = self.structure
        else:
            structure = self.structure.clone()
        for model in structure:
            for chain in model:
                for residue in chain:
                    for atom in residue:
                        atom.pos = gemmi.Position(
                            atom.pos.x + trans_vector[0],
                            atom.pos.y + trans_vector[1],
                            atom.pos.z + trans_vector[2],
                        )
                        if remove_charges:  # remove negative charges?
                            if atom.charge != 0:
                                atom.charge = 0
        if out_model_path:
            if os.path.splitext(out_model_path)[1].lower() in [".cif", ".mmcif"]:
                write_structure_as_mmcif(structure, out_model_path)
            else:
                write_structure_as_pdb(structure, out_model_path)
        return structure

    def remove_atomic_charges(
        self,
        out_model_path: Union[pathlib.Path, str] = None,
        inplace: bool = False,
    ):
        """
        Remove atomic charges from the model

        Arguments
        ---------
            :out_model_path:
                output atomic model with charges removed
        """
        if inplace:
            structure = self.structure
        else:
            structure = self.structure.clone()
        for model in structure:
            for chain in model:
                # skip non polymers
                # if not polymer: continue
                for residue in chain:
                    for atom in residue:
                        if atom.charge != 0:
                            atom.charge = 0
        if out_model_path:
            if os.path.splitext(out_model_path)[1].lower() in [".cif", ".mmcif"]:
                write_structure_as_mmcif(self.structure, out_model_path)
            else:
                write_structure_as_pdb(structure, out_model_path)
        return structure

    def set_residue_attribute(self, attr_name: str, dict_attribute: dict):
        """
        TODO: this is just a placeholder
        Add a new residue attribute to the current gemmi structure instance
        """
        for model in self.structure:
            if model.name not in dict_attribute:
                continue
            for chain in model:
                if chain.name in dict_attribute[model.name]:
                    for residue in chain:
                        pass
                        # try:
                        #     setattr(
                        #         residue,
                        #         attr_name,
                        #         dict_attribute[model.name][chain.name][
                        #             str(residue.seqid.num)
                        #         ],
                        #     )
                        # except KeyError:
                        #     pass

    def set_residue_types(self):
        """
        Set a dictionary to map residue IDs and names
        """
        self.dict_resnames = OrderedDict()
        for model in self.structure:
            self.dict_resnames[model.name] = {}
            for chain in model:
                if chain.name not in self.dict_resnames[model.name]:
                    self.dict_resnames[model.name][chain.name] = {}
                for residue in chain:
                    self.dict_resnames[model.name][chain.name][
                        str(residue.seqid.num)
                    ] = residue.name

    def get_avgbfact_deviation(
        self,
        calc_dev: bool = True,
        dist_dev: float = 5.0,
        save_obj: bool = False,
        out_json: str = None,
        skip_nonpoly: bool = True,
    ):
        """
        Get average atomic B-factors, and stdev of atomic B-factors
        around each residue in the model

        Returns
        -------
            Dict with B factors and  deviations
            {modelName:{chainName:
            {residueNum_residueName: [avg b-factor, b-factor dev]}}}
        """
        dict_bfact = OrderedDict()
        for model in self.structure:
            if calc_dev:
                subcells = gemmi.NeighborSearch(
                    model, self.structure.cell, dist_dev + 5.0
                )
                subcells.populate(include_h=False)
                dict_bfact[model.name] = {}
            for chain in model:
                polymer = chain.get_polymer()
                # skip non polymers
                if not polymer and skip_nonpoly:
                    continue
                if chain.name not in dict_bfact[model.name]:
                    dict_bfact[model.name][chain.name] = {}
                for residue in chain:
                    if residue.entity_type == gemmi.EntityType.Water:
                        continue
                    residue_id = str(residue.seqid.num)  # + "_" + residue.name
                    list_bfact = []
                    list_neigh_bfact = []
                    for atom in residue:
                        list_bfact.append(atom.b_iso)
                    avg_bfact = round(sum(list_bfact) / float(len(list_bfact)), 5)
                    if calc_dev:
                        representative_atom = self.get_representative_atom(
                            residue=residue
                        )
                        representative_atom_b = representative_atom.b_iso
                        marks = subcells.find_neighbors(
                            representative_atom, min_dist=0.1, max_dist=dist_dev
                        )
                        for mark in marks:
                            cra = mark.to_cra(model)
                            neigh_atom = cra.atom
                            list_neigh_bfact.append(neigh_atom.b_iso)
                        list_neigh_bfact.append(representative_atom_b)
                    if len(list_neigh_bfact) > 0:
                        dict_bfact[model.name][chain.name][residue_id] = [
                            avg_bfact,
                            round(np.std(np.array(list_neigh_bfact)), 5),
                        ]
                    else:
                        dict_bfact[model.name][chain.name][residue_id] = [
                            avg_bfact,
                            0.0,
                        ]

        # if save_obj:  # save gemmi structure object instead of returned dict
        #     self.set_residue_attribute(
        #         attr_name="bfact_avg_dev", dict_attribute=dict_bfact
        #     )
        #     if out_json is not None:
        #         encode_simple_obj_to_json_file(
        #             GemmiStructureExt(self.structure), out_json
        #         )
        #     else:
        #         encode_simple_obj_to_json_file(
        #             GemmiStructureExt(self.structure),
        #             self.modelid + "_gemmistructure.json",
        #         )
        #     print(dir(GemmiStructureExt(self.structure)))
        if out_json is not None:
            with open(out_json, "w") as oj:
                json.dump(dict_bfact, oj)
        return dict_bfact

    def get_coordinates(
        self,
        atom_selection: Union[str, list] = "all",
        skip_non_poly: bool = False,
        return_list: bool = False,
        out_json: str = None,
    ):
        """
        Get model atomic coordinates based on user selection

        Arguments
        ---------
            in_model_path: str
                input atomic model
            gemmi_structure_instance: gemmi.Structure, optional
                gemmi structure instance if in_model_path is not input
            atom_selection: Union[str, list], optional
                atom selection for coordinate retrieval
                Input a list of atom names or any of the following keywords:
                    "all": all atoms in the model
                    "backbone": only model backbone atoms, all atoms for non polymers
                    "one_per_residue": representative atoms, e.g. CA for amino acids
                    "centre": geometric centre of the residue
            skip_non_poly: bool, optional
                skip non polymers?
            return_list: bool, optional
                Returns a list of residue IDs and a list of coordinates
            out_json: str, optional
                json file to save the output coordinate dictionary
                (if return_list is False)
        Returns
        -------
            A dictionary of coordinates or list of residue IDs + a list of coordinates
            {modelName:{chainName: {residueNum_residueName: [list atomic coordinates]}}}
        """
        dict_coord = OrderedDict()
        list_ids = []
        list_coords = []
        polymertype = None
        # loop through structure
        for model in self.structure:
            if model.name not in dict_coord:
                dict_coord[model.name] = {}
            for chain in model:
                polymer = chain.get_polymer()
                # skip non polymers
                if not polymer and skip_non_poly:
                    continue
                if polymer.check_polymer_type() == gemmi.PolymerType.PeptideL:
                    polymertype = "protein"
                if chain.name not in dict_coord[model.name]:
                    dict_coord[model.name][chain.name] = {}
                for residue in chain:
                    list_residue_coords = []
                    if isinstance(atom_selection, list):  # specified atoms only
                        list_residue_coords = self.get_selected_atom_coordinates(
                            residue, atom_selection
                        )
                    elif atom_selection == "all":
                        for atom in residue:
                            list_residue_coords.append(
                                (atom.pos.x, atom.pos.y, atom.pos.z)
                            )
                    elif atom_selection == "backbone":  # all backbone
                        list_residue_coords = self.get_backbone_coordinates(
                            residue, polymertype=polymertype
                        )
                    elif atom_selection == "one_per_residue":  # ca or equivalent
                        residue_centre = self.get_representative_coordinate(
                            residue, polymertype=polymertype
                        )
                        list_residue_coords.append(residue_centre)
                    elif atom_selection == "centre":  # central atom
                        residue_centre = self.get_residue_centre_by_coordinates(residue)
                        list_residue_coords.append(residue_centre)
                    if return_list:
                        list_ids.extend(
                            len(list_residue_coords)
                            * [
                                "_".join(
                                    [
                                        model.name,
                                        str(chain.name),
                                        str(residue.seqid.num),
                                        residue.name,
                                    ]
                                )
                            ]
                        )
                        list_coords.extend(list_residue_coords)
                    else:
                        residue_coords = list_residue_coords
                        residue_id = str(residue.seqid.num)  # + "_" + residue.name
                        dict_coord[model.name][str(chain.name)][
                            residue_id
                        ] = residue_coords
        if not return_list:
            if out_json:
                with open(out_json, "w") as j:
                    json.dump(dict_coord, j)
            return dict_coord
        else:
            return list_ids, list_coords

    def get_selected_atom_coordinates(
        self, residue: gemmi.Residue, atom_selection: list
    ):
        """
        Get coordinates of selected atoms (list of atom names)
        from a given gemmi Residue instance
        """
        list_residue_coords = []
        for atom in residue:
            if atom.name in atom_selection:
                list_residue_coords.append((atom.pos.x, atom.pos.y, atom.pos.z))
        return list_residue_coords

    def get_representative_coordinate(
        self, residue: gemmi.Residue, polymertype: str = None
    ):
        """
        Get coordinates of one representative atom per residue
        from a given gemmi Residue instance
        """
        residue_centre = ()
        if residue.name in ["A", "T", "C", "G", "U"]:  # nuc acid
            for atom in residue:
                if atom.name in ["P", "C3'", "C1'"]:
                    residue_centre = (atom.pos.x, atom.pos.y, atom.pos.z)
        elif not polymertype or polymertype == "protein":  # protein
            for atom in residue:
                # if atom.name == "CB" or (residue.name == "GLY" and atom.name == "CA"):
                if atom.name == "CA":
                    residue_centre = (atom.pos.x, atom.pos.y, atom.pos.z)
                    break
        if len(residue_centre) == 0:  # non nuc acid / prot
            residue_centre = self.get_residue_centre_by_coordinates(residue)
        return residue_centre

    def get_representative_atom(self, residue: gemmi.Residue, polymertype: str = None):
        """
        Get one representative atom per residue
        from a given gemmi Residue instance
        """
        repr_atom = None
        if residue.name in ["A", "T", "C", "G", "U"]:  # nuc acid
            for atom in residue:
                if atom.name in ["P", "C3'", "C1'"]:
                    repr_atom = atom
        elif not polymertype or polymertype == "protein":  # protein
            for atom in residue:
                # if atom.name == "CB" or (residue.name == "GLY" and atom.name == "CA"):
                if atom.name == "CA":
                    repr_atom = atom
                    break
        if not repr_atom:  # non nuc acid / prot
            repr_atom = self.get_central_atom_by_sequence(residue)
        return repr_atom

    def get_backbone_coordinates(self, residue: gemmi.Residue, polymertype: str = None):
        """
        Get coordinates of the backbone atoms
        from a given gemmi Residue instance
        """
        list_residue_coords = []
        if residue.name in ["A", "T", "C", "G", "U"]:  # nuc acid
            for atom in residue:
                # TODO : fix the backbone atom list
                if atom.name in ["P", "C3'", "C1'"]:
                    list_residue_coords.append((atom.pos.x, atom.pos.y, atom.pos.z))
        elif residue.het_flag == "H":  # all het atoms
            for atom in residue:
                list_residue_coords.append((atom.pos.x, atom.pos.y, atom.pos.z))
        elif not polymertype or polymertype == "protein":  # protein
            for atom in residue:
                if atom.name in ["N", "CA", "C", "O"]:
                    list_residue_coords.append((atom.pos.x, atom.pos.y, atom.pos.z))
        if len(list_residue_coords) == 0:  # all
            for atom in residue:
                list_residue_coords.append((atom.pos.x, atom.pos.y, atom.pos.z))
        return list_residue_coords

    def get_residue_centre_by_coordinates(self, gemmi_residue_instance: gemmi.Residue):
        """
        Get coordinates of one central atom (sequence centre) per residue
        from a given gemmi Residue instance
        """
        list_coords = []
        for atom in gemmi_residue_instance:
            list_coords.append([atom.pos.x, atom.pos.y, atom.pos.z])
        return list(np.mean(np.array(list_coords), axis=0))

    def get_residue_centre_by_sequence(self, gemmi_residue_instance: gemmi.Residue):
        """
        Get coordinates of one central atom (sequence centre) per residue
        from a given gemmi Residue instance
        """
        try:
            center_index = int(len(gemmi_residue_instance) / 2)
            atom = gemmi_residue_instance[center_index]
            residue_centre = (atom.pos.x, atom.pos.y, atom.pos.z)
        except IndexError:
            for atom in gemmi_residue_instance:
                residue_centre = (atom.pos.x, atom.pos.y, atom.pos.z)
                break  # first atom
        return residue_centre

    def get_central_atom_by_sequence(self, gemmi_residue_instance: gemmi.Residue):
        """
        Get one central atom (sequence centre) per residue
        from a given gemmi Residue instance
        """
        try:
            center_index = int(len(gemmi_residue_instance) / 2)
            atom = gemmi_residue_instance[center_index]
            central_atom = atom
        except IndexError:
            for atom in gemmi_residue_instance:
                central_atom = atom
                break  # first atom
        return central_atom

    def get_sequence_from_atom_records(self):
        """
        Get polymer sequence from atom records
        """
        dict_seq = {}
        self.structure.add_entity_types(overwrite=True)
        self.structure.assign_subchains()  # assigns subchains in each chain
        self.structure.ensure_entities()  # each subchain linked to one Entity object
        for entity in self.structure.entities:
            if entity.entity_type == gemmi.EntityType.Polymer:
                seq_res_list = entity.full_sequence
                seq_res_list = [gemmi.Entity.first_mon(item) for item in seq_res_list]
                sequence = "".join(
                    [
                        gemmi.find_tabulated_residue(resname).one_letter_code
                        for resname in seq_res_list
                    ]
                )
                dict_seq[entity.name] = sequence
        return dict_seq

    def get_sequence_resnum_from_atom_records(self):
        """
        Get polymer sequence and list of residue number from model atom records

        Returns
        -------
            Dict of sequences of each chain
            Dict of residue numbers of each chain {'A': ['G 74', 'L 75', 'A 76', ....]}
        """
        dict_seq = {}
        dict_residues = {}
        for model in self.structure:
            for chain in model:
                polymer = chain.get_polymer()
                # skip non polymers
                if not polymer:
                    continue
                list_seq = []
                list_resnum = []
                for residue in chain:
                    rescode = gemmi.find_tabulated_residue(residue.name).one_letter_code
                    if rescode != " ":
                        list_seq.append(rescode)
                        list_resnum.append(rescode + " " + str(residue.seqid.num))
                sequence = "".join(list_seq)
                dict_seq[chain.name] = sequence
                dict_residues[chain.name] = list_resnum
        return dict_seq, dict_residues


def convert_pdb_to_mmcif(pdb_path, mmcif_path=None):
    """Convert coordinate file from PDB format to mmCIF/PDBx format"""
    structure = gemmi.read_structure(pdb_path)
    if mmcif_path is None:
        mmcif_path = os.path.splitext(pdb_path)[0] + ".cif"
    write_structure_as_mmcif(structure, mmcif_path)
    return mmcif_path


def write_structure_as_mmcif(structure, mmcif_name):
    """Write a Gemmi structure out to an mmCIF file."""
    # Refmac crashes for long _entry.id
    st_new = structure.clone()
    st_new.name = st_new.name[:78]  # in case of pdb this will be _entry.id
    if "_entry.id" in st_new.info:
        st_new.info["_entry.id"] = st_new.info["_entry.id"][:78]
    st_new.make_mmcif_document().write_file(mmcif_name)


def write_structure_as_pdb(structure, pdb_name):
    """Write a Gemmi structure out to a PDB file."""
    st_new = structure.clone()
    st_new.shorten_chain_names()
    st_new.write_pdb(pdb_name, use_linkr=True)


def convert_mmcif_to_pdb(mmcif_path, pdb_path=None):
    """Convert coordinate file from mmCIF/PDBx format to PDB format"""
    structure = gemmi.read_structure(mmcif_path)
    if pdb_path is None:
        pdb_path = os.path.splitext(pdb_path)[0] + ".pdb"
    write_structure_as_pdb(structure, pdb_path)
    return pdb_path


def check_structure_inputs(input_structure: Union[str, pathlib.Path, gemmi.Structure]):
    """
    Check if the given input is a model file path or a gemmi Structure instance

    Returns
    -------
    gemmi Structure instance
    """
    if isinstance(input_structure, gemmi.Structure):
        structure = input_structure
    elif input_structure is not None and os.path.isfile(input_structure):
        structure = gemmi.read_structure(input_structure)
    else:
        raise IOError("Input either a valid model file or a gemmi structure instance")
    return structure


def get_residue_attribute(list_ids: List[str], list_attr: List[Union[float, int]]):
    """
    Get residue attributes as average of atom attributes

    Arguments
    ---------
        :list_ids:
            list of residue or atom IDs of the format:
            residue_id = "_".join(
                    [
                        model name,
                        chain name,
                        str(residue number),
                        residue name,
                    ]
                )
            atom_id = "_".join([residue_id, atom name])
        :list_attr:
            list of atom attributes (same length as list_ids)

    Returns:
        Lists of residue IDs and average attributes
    """
    list_res_ids = []
    list_res_attr = []
    dict_current_res = {}
    # model.name_chain.name_residue.seqid.num_residue.name
    atom_number = 0
    previous_res_id = None
    for atom_id in list_ids:
        atom_id_split = atom_id.split("_")
        if len(atom_id_split) > 4:  # has atom identifier
            res_id = "_".join(atom_id_split[:4])
        else:
            res_id = atom_id
        try:
            dict_current_res[res_id].append(list_attr[atom_number])
        except KeyError:
            if previous_res_id:
                list_res_ids.append(previous_res_id)
                if len(dict_current_res[previous_res_id]) == 1:
                    list_res_attr.append(dict_current_res[previous_res_id][0])
                elif len(dict_current_res[previous_res_id]) > 1:
                    list_res_attr.append(
                        round(
                            np.sum(dict_current_res[previous_res_id])
                            / len(dict_current_res[previous_res_id]),
                            5,
                        )
                    )
            dict_current_res = {}  # reinitialize
            dict_current_res[res_id] = [list_attr[atom_number]]
        previous_res_id = res_id
        atom_number += 1
    if previous_res_id:
        list_res_ids.append(previous_res_id)
        if len(dict_current_res[previous_res_id]) == 1:
            list_res_attr.append(dict_current_res[previous_res_id][0])
        elif len(dict_current_res[previous_res_id]) > 1:
            list_res_attr.append(
                round(
                    np.sum(dict_current_res[previous_res_id])
                    / len(dict_current_res[previous_res_id]),
                    5,
                )
            )
    return list_res_ids, list_res_attr


def convert_dict_attributes_to_list(dict_attr):
    list_ids = []
    list_attr = []
    for model in dict_attr:
        for chain in dict_attr[model]:
            for residue in dict_attr[model][chain]:
                if isinstance(dict_attr[model][chain][residue], dict):
                    for atom in dict_attr[model][chain][residue]:
                        atom_id = "_".join([model, str(chain), str(residue), str(atom)])
                        list_ids.extend(
                            len(dict_attr[model][chain][residue][atom]) * [atom_id]
                        )
                        list_attr.extend(dict_attr[model][chain][residue][atom])

                else:
                    res_id = "_".join([model, str(chain), str(residue)])
                    list_ids.extend(len(dict_attr[model][chain][residue]) * [res_id])
                    list_attr.extend(dict_attr[model][chain][residue])
    return list_ids, list_attr


def set_bfactor_attributes(
    in_model_path: Union[str, pathlib.Path],
    dict_attr: dict,
    skip_non_poly: bool = True,
    attr_name: str = "",
    out_dir: str = None,
):
    """
    Replace atomic B-factors with the given attributes
    Use this for visualisation purpose (colr by b-factors) only

    Arguments
    ---------
        in_model_path:
            input atomic model path
        dict_attr:
            Dictionary of residue or atom attributes
            Dict keys are residue or atom IDs of the format:
            residue_id = "_".join(
                    [
                        model name,
                        chain name,
                        str(residue number),
                        residue name,
                    ]
                )
            atom_id = "_".join([residue_id, atom name])
        skip_non_poly: bool, optional
            skip non polymers?
    Returns
    -------
        A dictionary of coordinates or list of residue IDs + a list of coordinates

    """
    structure = gemmi.read_structure(in_model_path)
    # setup entities in case that nothing were set when reading the file
    structure.setup_entities()
    structure.assign_label_seq_id()
    count_atom = 0
    for model in structure:
        for chain in model:
            polymer = chain.get_polymer()
            # skip non polymers
            if not polymer and skip_non_poly:
                continue
            for residue in chain:
                residue_id = "_".join(
                    [
                        model.name,
                        str(chain.name),
                        str(residue.seqid.num),
                        residue.name,
                    ]
                )
                for atom in residue:
                    atom_id = "_".join([residue_id, atom.name])
                    if atom_id in dict_attr:
                        atom.b_iso = dict_attr[atom_id]
                    elif residue_id in dict_attr:
                        atom.b_iso = dict_attr[residue_id]
                    else:
                        atom.b_iso = 0.0
                    count_atom += 1
    if out_dir:
        structure.write_pdb(
            os.path.join(
                out_dir,
                os.path.splitext(os.path.basename(in_model_path))[0]
                + "_"
                + attr_name
                + "_attr.pdb",
            )
        )
    else:
        structure.write_pdb(
            os.path.splitext(os.path.basename(in_model_path))[0]
            + "_"
            + attr_name
            + "_attr.pdb"
        )



import numpy as np
from typing import Union, Type

try:
    from scipy.spatial import cKDTree as kdt
except ImportError:
    from scipy.spatial import KDTree as kdt

import math
import re


def calc_z(
    list_vals: Union[list, np.ndarray],
    val: float = None,
) -> Union[float, list]:
    # remove selected value before calculating z
    if val is not None:
        if isinstance(val, list):
            list_vals.remove(val)  # remove val from list
        elif isinstance(val, np.ndarray):
            list_vals = list_vals[list_vals != val]
        else:
            pass
    sum_mapval = np.sum(list_vals)
    mean_mapval = sum_mapval / len(list_vals)
    # median_mapval = np.median(list_vals)
    std_mapval = np.std(list_vals)
    if val is None:
        if std_mapval == 0.0:
            z = [0.0] * len(list_vals)
        else:
            z = np.round((np.array(list_vals) - mean_mapval) / std_mapval, 2)
    else:
        if std_mapval == 0.0:
            z = 0.0
        else:
            z = (val - mean_mapval) / std_mapval
    return z


def calc_z_median(
    list_vals: Union[list, np.ndarray],
    val: float = None,
) -> Union[float, list]:
    # remove selected value before calculating z
    if val is not None:
        if isinstance(val, list):
            list_vals.remove(val)  # remove val from list
        elif isinstance(val, np.ndarray):
            list_vals = list_vals[list_vals != val]
        else:
            pass

    # sum_smoc = np.sum(list_vals)
    median_smoc = np.median(list_vals)
    mad_smoc = np.median(np.absolute(np.array(list_vals) - median_smoc))
    if val is None:
        if mad_smoc == 0.0:
            z = [0.0] * len(list_vals)
        else:
            z = np.around((np.array(list_vals) - median_smoc) / mad_smoc, 2)
    else:
        if mad_smoc == 0.0:
            z = 0.0
        else:
            z = round(((val - median_smoc) / mad_smoc), 2)
    return z


def get_indices_sphere(
    gridtree: Type[kdt],
    coord: Union[list, np.ndarray],
    dist: float = 5.0,
) -> list:
    list_points = gridtree.query_ball_point([coord[0], coord[1], coord[2]], dist)
    return list_points


# utility functions to compare dict/list/str/num including a tolerance
# for any numeric comparisons within
def compare_dict_almost_equal(self, dict1: dict, dict2: dict, num_rel_tol: float):
    diff_dict = {}
    for k in dict2:
        assert k in dict1
        if dict1[k] != dict2[k]:
            diff_dict[k] = dict2[k]
    for k in diff_dict:
        if isinstance(dict1[k], dict):
            self.compare_dict_similarity(dict1[k], dict2[k], num_rel_tol)
        elif type(dict1[k]) in [list, tuple]:
            self.compare_list_similarity(dict1[k], dict2[k], num_rel_tol)
        else:
            self.compare_values(dict1[k], dict2[k], num_rel_tol)


def compare_list_almost_equal(self, list1, list2, num_rel_tol):
    for n in range(len(list1)):
        if isinstance(list1[n], dict):
            self.compare_dict_similarity(list1[n], list2[n], num_rel_tol)
        if type(list1[n]) in [list, tuple]:
            self.compare_list_similarity(list1[n], list2[n], num_rel_tol)
        else:
            self.compare_values(list1[n], list2[n], num_rel_tol)


def compare_values_almost_equal(self, value1, value2, num_rel_tol):
    if isinstance(value1, str):
        l1 = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", value1)
        l2 = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", value2)
        assert len(l1) == len(l2)
        if len(l1) == 1:
            assert math.isclose(float(l1[0]), float(l2[0]), rel_tol=num_rel_tol)
        else:
            assert l1 == l2
    else:
        try:
            assert math.isclose(float(l1), float(l2), rel_tol=num_rel_tol)
        except ValueError:
            assert l1 == l2


def get_moc(array1: np.ndarray, array2: np.ndarray) -> float:
    num = np.sum(array1 * array2)
    den = np.sqrt(np.sum(np.square(array1)) * np.sum(np.square(array2)))
    if den == 0.0:
        if num == 0.0:
            return 1.0
        return -1.0
    return num / den


def get_ccc(array1: np.ndarray, array2: np.ndarray) -> float:
    num = np.sum((array1 - np.mean(array1)) * (array2 - np.mean(array2)))
    den = np.sqrt(
        np.sum(np.square(array1 - np.mean(array1)))
        * np.sum(np.square(array2 - np.mean(array2)))
    )
    if den == 0.0:
        if num == 0.0:
            return 1.0
        return -1.0
    return num / den



import numpy as np
import math
from scipy.ndimage import generic_filter, measurements, gaussian_filter
from typing import Sequence
from scipy.ndimage import binary_opening, binary_closing
import gc

try:
    import pyfftw

    pyfftw_flag = True
except ImportError:
    pyfftw_flag = False


def plot_density_histogram(arr, plotfile="density_histogram.png"):
    freq, bins = np.histogram(arr, 100)
    bin_centre = []
    for i in range(len(bins) - 1):
        bin_centre.append((bins[i] + bins[i + 1]) / 2.0)
    assert len(freq) == len(bin_centre)

    try:
        import matplotlib.pyplot as plt
    except (RuntimeError, ImportError) as e:
        if hasattr(e, "message"):
            print("Matplotlib import error {}".format(e.message))
        else:
            print(e)
        plt = None
        return
    import warnings

    warnings.filterwarnings("ignore", category=RuntimeWarning)
    logfreq = []
    for f in freq:
        try:
            logfreq.append(np.log(f))
        except RuntimeWarning:
            logfreq.append(0.0)
    logfreq = np.array(logfreq)
    logfreq[freq == 0] = 0.0
    plt.rcParams.update({"font.size": 15})
    plt.xlabel("Density", fontsize=12)
    plt.ylabel("log(Frequency)", fontsize=12)
    plt.plot(bin_centre, logfreq, linewidth=2.0, color="g")
    #     locs,labs = plt.xticks()
    #     step = (max(locs)-min(locs))/10.
    #     locs = np.arange(min(locs),max(locs)+step,step)
    #     labels = np.round(map_apix/locs[1:],1)
    #     plt.xticks(locs[1:], labels,rotation='vertical')
    plt.savefig(plotfile)
    plt.close()


def get_contour_mask(
    array: np.ndarray,
    threshold_level: float,
) -> np.ndarray:
    binary_mask = array > float(threshold_level)
    return binary_mask


def threshold_array(
    array: np.ndarray,
    threshold_level: float,
    dtype="float32",
    inplace=False,
) -> np.ndarray:
    binary_mask = get_contour_mask(array, threshold_level)
    return appy_mask(array, binary_mask, dtype=dtype, inplace=inplace)


def appy_mask(
    arr: np.ndarray,
    mask: np.ndarray,
    dtype: str = "float32",
    inplace: bool = False,
):
    if not compare_tuple(arr.shape, mask.shape):
        return arr
    if inplace:
        arr[:] = arr * mask
        return arr
    else:
        return np.array(arr * mask, dtype=dtype)


def shift_values(arr, offset, inplace=False):
    """
    Shift density given an offset
    """
    if inplace:
        arr[:] = arr + float(offset)
        return arr
    else:
        return arr + float(offset)


def crop_array(
    arr: np.ndarray,
    new_dim: Sequence[int] = None,
    contour: float = None,
    ext: int = None,
    cubic: bool = False,
    inplace: bool = False,
    mask_array: np.ndarray = None,
    nd: int = 3,
) -> np.ndarray:
    """
    Crop an array based on
        1) an array value threshold or mask
            (plus optional
                padding (ext) applied as number of pixels/voxels)
        2) new array dimensions

    Arguments
    ---------
        :arr: np.ndarray
            input array (indexed as arr[z][y][x])
        :new_dim:
            new dimensions to crop to
        :contour: float, optional
            map threshold (slices with all values lower than this are cropped)
        :ext: int, optional
            number of pixels/voxels as padding
        :mask_array: np.ndarray
            mask to be applied before cropping (slices with all zeros are cropped)
        :cubic: bool
            whether to crop to a cubic box?

    Returns
    -------
        Cropped array
    """
    # crop based on given dimensions
    if new_dim is not None:
        crop_x = int(arr.shape[2] - new_dim[0])
        xs = max(0, int(np.ceil(float(crop_x) / 2)))
        xe = min(
            arr.shape[2],
            arr.shape[2] - int(np.floor(float(crop_x) / 2)),
        )

        crop_y = int(arr.shape[1] - new_dim[1])
        ys = max(0, int(np.ceil(float(crop_y) / 2)))
        ye = min(
            arr.shape[1],
            arr.shape[1] - int(np.floor(float(crop_y) / 2)),
        )

        crop_z = int(arr.shape[0] - new_dim[2])
        zs = max(0, int(np.ceil(float(crop_z) / 2)))
        ze = min(
            arr.shape[0],
            arr.shape[0] - int(np.floor(float(crop_z) / 2)),
        )
    # crop based on the give n contour and factor_sigma
    elif contour or mask_array:
        if contour:
            minval = float(contour)
        elif mask_array:
            assert mask_array.shape == arr.shape
            arr = arr * mask_array  # apply mask
            minval = 0.0000000001

        map_data = arr
        list_indices = []
        for i in range(nd):
            ct1 = 0
            try:
                while np.nanmax(map_data[ct1]) < minval:

                    ct1 += 1
            except IndexError:
                pass

            ct2 = 0
            try:
                while np.nanmax(map_data[-1 - ct2]) < minval:
                    ct2 += 1
            except IndexError:
                pass
            # transpose
            map_data = np.transpose(map_data, (2, 0, 1))
            # TODO, substracting 1 is not necessary?
            list_indices.append([ct1 - 1, ct2 - 1])

        # indices for cropping
        # z axis
        zs, ze = max(0, list_indices[0][0] - ext), min(
            arr.shape[0] - list_indices[0][1] + ext, arr.shape[0]
        )
        # y axis
        ys, ye = max(0, list_indices[2][0] - ext), min(
            arr.shape[1] - list_indices[2][1] + ext, arr.shape[1]
        )
        # x axis
        xs, xe = max(0, list_indices[1][0] - ext), min(
            arr.shape[2] - list_indices[1][1] + ext, arr.shape[2]
        )

        # make cubic dimensions
        if cubic:
            s_min = min([zs, ys, xs])
            e_max = max([ze, ye, xe])
            zs = s_min
            xs = s_min
            ys = s_min

            ze = e_max
            ye = e_max
            xe = e_max

        # cropped data, save a copy to get a contiguous memory block
        # delete the reference
        del map_data
    if inplace:
        return arr[zs:ze, ys:ye, xs:xe], (xs, ys, zs)
    else:
        return np.copy(arr[zs:ze, ys:ye, xs:xe]), (xs, ys, zs)


def pad_array(arr, nx, ny, nz):
    """

    Pad an array with specified increments along each dimension.
    Arguments:
        *nx,ny,nz*
           Number of slices to add to each dimension.
    Return:
        array
    """

    gridshape = (arr.shape[0] + nz, arr.shape[1] + ny, arr.shape[2] + nx)
    input_dtype = str(arr.dtype)
    new_array = np.zeros(gridshape, dtype=input_dtype)
    # min fill
    new_array.fill(arr.min())
    #
    oldshape = arr.shape
    indz, indy, indx = (
        int(round((gridshape[0] - oldshape[0]) / 2.0)),
        int(round((gridshape[1] - oldshape[1]) / 2.0)),
        int(round((gridshape[2] - oldshape[2]) / 2.0)),
    )
    # copy the data
    new_array[
        indz : indz + oldshape[0], indy : indy + oldshape[1], indx : indx + oldshape[2]
    ][:] = arr
    return new_array, (indx, indy, indz)


# footprint array corresponding to 6 neighboring faces of a voxel
def grid_footprint():
    """
    Generate a footprint array for local neighborhood (6 faces)
    """
    a = np.zeros((3, 3, 3))
    a[1, 1, 1] = 1
    a[0, 1, 1] = 1
    a[1, 0, 1] = 1
    a[1, 1, 0] = 1
    a[2, 1, 1] = 1
    a[1, 2, 1] = 1
    a[1, 1, 2] = 1

    return a


# SMOOTHING/DUSTING
# useful when a 'safe' contour level is chosen
# for high resolution maps with fine features on the surface,
# this might erode part of useful density
def map_binary_opening(arr, contour, it=1, inplace=False):
    """
    Remove isolated densities by erosion
    """
    fp = grid_footprint()
    # current position can be updated based on neighbors only
    fp[1, 1, 1] = 0
    if inplace:
        arr[:] = arr * binary_opening(arr > float(contour), structure=fp, iterations=it)
        return arr
    else:
        return arr * binary_opening(arr > float(contour), structure=fp, iterations=it)


def map_binary_closing(arr, contour, it=1, inplace=False):
    """
    Close/Fill 'empty' (zero) voxels with filled (non-zero) neighbors
    """
    fp = grid_footprint()
    # current position can be updated based on neighbors only
    fp[1, 1, 1] = 0
    # threshold can be 1*std() to be safe?
    if inplace:
        arr[:] = arr * binary_closing(arr > float(contour), structure=fp, iterations=it)
        return arr
    else:
        return arr * binary_closing(arr > float(contour), structure=fp, iterations=it)


def apply_filter(ftmap, ftfilter, inplace=False):
    # fftshifted ftmap
    if inplace:
        ftmap[:] = ftmap * ftfilter
        return ftmap
    else:
        return ftmap * ftfilter


def find_level(arr, vol, apix):
    """
    Get the threshold corresponding to volume.
    """

    # initiate with reasonable values
    c1 = arr.min()
    vol_calc = float(vol) * 2
    # loop until calc vol and exp vol matches
    it = 0
    flage = 0
    while (vol_calc - float(vol)) / (apix[0] * apix[1] * apix[2]) > 10 and flage == 0:
        # mask only values >= previous sel
        mask_array = arr >= c1
        # compute histogram wt 1000 bins
        dens_freq, dens_bin = np.histogram(arr[mask_array], 1000)
        # loop over bins to select a min level from the bins
        sum_freq = 0.0
        for i in range(len(dens_freq)):
            sum_freq += dens_freq[-(i + 1)]
            dens_level = dens_bin[-(i + 2)]
            vol_calc = sum_freq * (apix[0] * apix[1] * apix[2])
            # break when vol exceeds exp vol to select the bin level
            if vol_calc > float(vol):
                sel_level = dens_level
                it += 1
                if sel_level <= c1:
                    flage = 1
                # print it, sel_level, c1, sum_freq, vol_calc, vol, flage
                c1 = sel_level
                if it == 3:
                    flage = 1
                break
    return sel_level


# sizes of all patches
def size_patches(arr, contour):
    """
    Get sizes or size distribution of isolated densities
        Arguments:
            *contour*
                density threshold
        Return:
            array of sizes
    """
    fp = grid_footprint()
    binmap = arr > float(contour)
    label_array, labels = measurements.label(arr * binmap, structure=fp)
    sizes = measurements.sum(binmap, label_array, range(labels + 1))
    return sizes


def get_sigma_map(arr, window=5):
    footprint_sph = make_spherical_footprint(window)
    sigma_array = generic_filter(
        arr, np.std, footprint=footprint_sph, mode="constant", cval=0.0
    )
    return sigma_array


def get_fsc(map1_array, map2_array, map_apix=1.0, maxRes=None, keep_shape=False):
    dist1 = make_fourier_shell(
        map1_array.shape, keep_shape=keep_shape, fftshift=False, normalise=False
    )
    dist1 = dist1.astype(np.int)
    # dist2 = make_fourier_shell(map2_array.shape,keep_shape=keep_shape,
    #                           fftshift=False,normalise=False)
    # dist2 = dist2.astype(np.int)
    ftarray1 = calculate_fft(map1_array, keep_shape=keep_shape)
    ftarray2 = calculate_fft(map2_array, keep_shape=keep_shape)

    ftarray1 = np.fft.fftshift(ftarray1, axes=(0, 1))
    ftarray2 = np.fft.fftshift(ftarray2, axes=(0, 1))

    listfreq, listfsc, listnsf = calculate_fsc(
        ftarray1, ftarray2, dist1, map_apix=map_apix
    )

    return calculate_fscavg(listfreq, listfsc, listnsf, maxRes=maxRes)


def get_fsc_curve(map1_array, map2_array, map_apix=1.0, keep_shape=False):
    dist1 = make_fourier_shell(
        map1_array.shape, keep_shape=keep_shape, fftshift=False, normalise=False
    )
    dist1 = dist1.astype(np.int)
    # dist2 = make_fourier_shell(map2_array.shape,keep_shape=keep_shape,
    #                           fftshift=False,normalise=False)
    # dist2 = dist2.astype(np.int)
    ftarray1 = calculate_fft(map1_array, keep_shape=keep_shape)
    ftarray2 = calculate_fft(map2_array, keep_shape=keep_shape)

    ftarray1 = np.fft.fftshift(ftarray1, axes=(0, 1))
    ftarray2 = np.fft.fftshift(ftarray2, axes=(0, 1))

    listfreq, listfsc, listnsf = calculate_fsc(
        ftarray1, ftarray2, dist1, map_apix=map_apix
    )

    return listfreq, listfsc


def get_raps(arr, keep_shape=False, plotfile="ps.png", map_apix=1.0):
    # make frequency shells
    dist = make_fourier_shell(
        arr.shape, keep_shape=keep_shape, fftshift=False, normalise=False
    )
    dist = dist.astype(np.int)
    ftarray = calculate_fft(arr, keep_shape=keep_shape)
    list_freq = []
    list_intensities = []
    for r in np.unique(dist)[0 : int(np.ceil(arr.shape[0] / 2))]:
        shell_indices = dist == r
        shell_amplitude = np.absolute(ftarray[shell_indices])
        avg_shell_intensity = np.log10(np.mean(np.square(shell_amplitude)))
        list_freq.append(float(r) / arr.shape[0])
        list_intensities.append(avg_shell_intensity)
    try:
        import matplotlib.pyplot as plt
    except (RuntimeError, ImportError) as e:
        if hasattr(e, "message"):
            print("Matplotlib import error {}".format(e.message))
        else:
            print(e)
        plt = None
        return list_freq, list_intensities

    plt.rcParams.update({"font.size": 12})
    plt.xlabel("Resolution", fontsize=12)
    plt.ylabel("log(Intensity)", fontsize=12)
    plt.plot(list_freq, list_intensities, linewidth=2.0, color="g")
    locs, labs = plt.xticks()
    step = (max(locs) - min(locs)) / 10.0
    locs = np.arange(min(locs), max(locs) + step, step)
    labels = np.round(map_apix / locs[1:], 1)
    plt.xticks(locs[1:], labels, rotation="vertical")
    plt.savefig(plotfile)
    plt.close()
    return list_freq, list_intensities


def calculate_fscavg(
    listfreq, listfsc, listnsf, maxRes=None, minRes=None, list_weights=None
):
    if maxRes is None:
        maxfreq_index = len(listfreq)
    else:
        maxfreq = 1.0 / maxRes
        # print maxfreq
        maxfreq_index = np.searchsorted(listfreq, maxfreq, side="right")
    if minRes is None:
        minfreq_index = 0
    else:
        minfreq = 1.0 / minRes
        minfreq_index = np.searchsorted(listfreq, minfreq)
    maxfreq_index = min(maxfreq_index, len(listfreq))

    weighted_fsc_sum = 0.0
    sum_nsf = 0.0
    for ind in range(minfreq_index, maxfreq_index):
        weighted_fsc_sum += listfsc[ind] * listnsf[ind]
        sum_nsf += listnsf[ind]
    FSCavg = weighted_fsc_sum / sum_nsf
    return FSCavg


def find_background_peak(arr, iter=2):
    lbin = np.amin(arr)
    rbin = np.amax(arr)
    ave = np.mean(arr)
    sigma = np.std(arr)
    for it in range(iter):
        if it == 0:
            data = arr
        else:
            data = arr[(arr >= lbin) & (arr <= rbin)]

        freq, bins = np.histogram(data, 100)
        ind = np.nonzero(freq == np.amax(freq))[0]
        peak = None
        for i in ind:
            val = (bins[i] + bins[i + 1]) / 2.0
            if val < float(ave) + float(sigma):
                peak = val
                lbin = bins[i]
                rbin = bins[i + 1]
        if peak is None:
            break
    return peak, ave


def plan_fft(arr, keep_shape=False, new_inparray=False):
    input_dtype = str(arr.dtype)
    if input_dtype not in ["float32", "float64", "longdouble"]:
        input_dtype = "float32"
        arr = arr.astype("float32")
    if not keep_shape:
        output_dtype = "complex64"
        if input_dtype == "float64":
            output_dtype = "complex128"
        elif input_dtype == "longdouble":
            output_dtype = "clongdouble"
        # for r2c transforms:
        output_array_shape = arr.shape[: len(arr.shape) - 1] + (arr.shape[-1] // 2 + 1,)
    else:
        output_dtype = "complex64"
        output_array_shape = arr.shape

    fftoutput = pyfftw.empty_aligned(output_array_shape, n=16, dtype=output_dtype)
    # check if array is byte aligned
    # TODO: can we read the map file as byte aligned?
    if new_inparray or not pyfftw.is_byte_aligned(arr):
        inputarray = pyfftw.empty_aligned(arr.shape, n=16, dtype=input_dtype)
        fft = pyfftw.FFTW(
            inputarray,
            fftoutput,
            direction="FFTW_FORWARD",
            axes=(0, 1, 2),
            flags=["FFTW_ESTIMATE"],
        )
    elif pyfftw.is_byte_aligned(arr):
        fft = pyfftw.FFTW(
            arr,
            fftoutput,
            direction="FFTW_FORWARD",
            axes=(0, 1, 2),
            flags=["FFTW_ESTIMATE"],
        )
        inputarray = np.copy(arr)

    return fft, fftoutput, inputarray


def calculate_fft(arr, keep_shape=False, new_inparray=False):

    if pyfftw_flag:
        fft, fftoutput, inputarray = plan_fft(
            arr, keep_shape=keep_shape, new_inparray=new_inparray
        )
        inputarray[:, :, :] = arr
        fft()
    else:
        # TODO: warning raises error in tasks
        # warnings.warn("PyFFTw not found!, using numpy fft")
        if not keep_shape:
            fftoutput = np.fft.rfftn(arr)
        else:
            fftoutput = np.fft.fftn(arr)
    return fftoutput


def plan_ifft(arr, output_shape=None, output_array_dtype=None, new_inparray=False):
    input_dtype = str(arr.dtype)
    #         #for c2r transforms:
    #             if output_shape is None: output_shape = \
    #                                     arr.shape[:len(arr.shape)-1]+\
    #                                     ((arr.shape[-1] - 1)*2,)
    if output_array_dtype is None:
        output_array_dtype = "float32"
    if output_shape is None:
        output_shape = arr.shape[: len(arr.shape) - 1] + ((arr.shape[-1] - 1) * 2,)
        if input_dtype not in ["complex64", "complex128", "clongdouble"]:
            input_dtype = "complex64"
        elif input_dtype == "complex128":
            output_array_dtype = "float64"
        elif input_dtype == "clongdouble":
            output_array_dtype = "longdouble"
    elif output_shape[-1] // 2 + 1 == arr.shape[-1]:
        if input_dtype not in ["complex64", "complex128", "clongdouble"]:
            input_dtype = "complex64"
        elif input_dtype == "complex128":
            output_array_dtype = "float64"
        elif input_dtype == "clongdouble":
            output_array_dtype = "longdouble"
    else:
        output_shape = arr.shape
        output_array_dtype = "complex64"

    output_array = pyfftw.empty_aligned(output_shape, n=16, dtype=output_array_dtype)
    # check if array is byte aligned
    if new_inparray or not pyfftw.is_byte_aligned(arr):
        inputarray = pyfftw.n_byte_align_empty(arr.shape, n=16, dtype=input_dtype)
        ifft = pyfftw.FFTW(
            inputarray,
            output_array,
            direction="FFTW_BACKWARD",
            axes=(0, 1, 2),
            flags=["FFTW_ESTIMATE"],
        )  # planning_timelimit=0.5)
        inputarray[:, :, :] = arr
    else:
        ifft = pyfftw.FFTW(
            arr,
            output_array,
            direction="FFTW_BACKWARD",
            axes=(0, 1, 2),
            flags=["FFTW_ESTIMATE"],
        )  # planning_timelimit=0.5)
        inputarray = arr

    return ifft, output_array, inputarray


def calculate_ifft(arr, output_shape=None, inplace=False, new_inparray=False):
    """
    Calculate inverse fourier transform
    """
    if pyfftw_flag:
        ifft, output_array, inputarray = plan_ifft(
            arr, output_shape=output_shape, new_inparray=new_inparray
        )
        # r2c fft
        ifft()
    else:
        # TODO: warnings raises error in tasks
        # warnings.warn("PyFFTw not found!, using numpy fft")
        if output_shape is None or output_shape[-1] // 2 + 1 == arr.shape[-1]:
            output_array = np.fft.irfftn(arr, s=output_shape)
        else:
            output_array = np.real(np.fft.ifftn(arr))
        # np.fft.fftpack._fft_cache.clear()
    del arr
    return output_array.real.astype(np.float32, copy=False)


def make_fourier_shell(map_shape, keep_shape=False, normalise=True, fftshift=True):
    """
    For a given grid, make a grid with sampling
    frequencies in the range (0:0.5)
    Return:
        grid with sampling frequencies
    """
    z, y, x = map_shape
    # numpy fftfreq : odd-> 0 to (n-1)/2 & -1 to -(n-1)/2 and
    # even-> 0 to n/2-1 & -1 to -n/2 to check with eman
    # set frequencies in the range -0.5 to 0.5
    if keep_shape:
        rad_z, rad_y, rad_x = np.mgrid[
            -np.floor(z / 2.0) : np.ceil(z / 2.0),
            -np.floor(y / 2.0) : np.ceil(y / 2.0),
            -np.floor(x / 2.0) : np.ceil(x / 2.0),
        ]
        if not fftshift:
            rad_x = np.fft.ifftshift(rad_x)
    # r2c arrays from fftw/numpy rfft
    else:
        rad_z, rad_y, rad_x = np.mgrid[
            -np.floor(z / 2.0) : np.ceil(z / 2.0),
            -np.floor(y / 2.0) : np.ceil(y / 2.0),
            0 : np.floor(x / 2.0) + 1,
        ]
    if not fftshift:
        rad_z = np.fft.ifftshift(rad_z)
        rad_y = np.fft.ifftshift(rad_y)
    if normalise:
        rad_z = rad_z / float(np.floor(z))
        rad_y = rad_y / float(np.floor(y))
        rad_x = rad_x / float(np.floor(x))
    rad_x = rad_x**2
    rad_y = rad_y**2
    rad_z = rad_z**2
    dist = np.sqrt(rad_z + rad_y + rad_x)
    return dist


def tanh_lowpass(map_shape, cutoff, fall=0.3, keep_shape=False):
    """
    Lowpass filter with a hyperbolic tangent function

    cutoff: high frequency cutoff [0:0.5]
    fall: smoothness of falloff [0-> tophat,1.0-> gaussian]
    Return:
        tanh lowpass filter to apply on fourier map
    """
    # e.g cutoff = apix/reso is the stop band
    if fall == 0.0:
        fall = 0.01
    drop = math.pi / (2 * float(cutoff) * float(fall))
    cutoff = min(float(cutoff), 0.5)
    # fall determines smoothness of falloff, 0-> tophat, 1.0-> gaussian
    # make frequency shells
    dist = make_fourier_shell(map_shape, keep_shape=keep_shape, fftshift=True)
    # filter
    dist1 = dist + cutoff
    dist1[:] = drop * dist1
    dist1[:] = np.tanh(dist1)
    dist[:] = dist - cutoff
    dist[:] = drop * dist
    dist[:] = np.tanh(dist)
    dist[:] = dist1 - dist
    dist = 0.5 * dist
    del dist1

    #     list_freq = []
    #     list_intensities = []
    #     rprev = 0.0
    #     for r in np.arange(0.02,0.5,0.02):
    #         shell_indices = (dist_ini < r) & (dist_ini <= rprev)
    #         avg_shell_intensity = np.mean(dist[shell_indices])
    #         list_freq.append(r)
    #         list_intensities.append(avg_shell_intensity)
    #         rprev = r
    #     print list_freq
    #     print list_intensities
    #
    return dist


def tanh_bandpass(
    map_shape,
    low_cutoff=0.0,
    high_cutoff=0.5,
    low_fall=0.1,
    high_fall=0.1,
    keep_shape=False,
):
    """
    Bandpass filter with a hyperbolic tangent function
    low_cutoff: low frequency cutoff [0:0.5]
    high-cutoff : high frequency cutoff [0:0.5]
    fall: determines smoothness of falloff [0-> tophat,1.0-> gaussian]
    Return:
        tanh lowpass filter to apply on fourier map
    """
    low_drop = math.pi / (2 * float(high_cutoff - low_cutoff) * float(low_fall))
    high_drop = math.pi / (2 * float(high_cutoff - low_cutoff) * float(high_fall))

    dist = make_fourier_shell(map_shape, keep_shape=keep_shape, fftshift=True)
    return 0.5 * (
        np.tanh(high_drop * (dist + high_cutoff))
        - np.tanh(high_drop * (dist - high_cutoff))
        - np.tanh(low_drop * (dist + low_cutoff))
        + np.tanh(low_drop * (dist - low_cutoff))
    )


def butterworth_lowpass(map_shape, pass_freq, keep_shape=False):
    """
    Lowpass filter with a gaussian function
    pass_freq : low-pass cutoff frequency [0:0.5]
    """
    eps = 0.882
    a = 10.624
    high_cutoff = (
        0.15 * math.log10(1.0 / pass_freq) + pass_freq
    )  # stop band frequency (used to determine the fall off)

    fall = 2.0 * (
        math.log10(eps / math.sqrt(a**2 - 1))
        / math.log10(pass_freq / float(high_cutoff))
    )

    cutoff_freq = float(pass_freq) / math.pow(eps, 2 / fall)

    dist = make_fourier_shell(map_shape, keep_shape=keep_shape, fftshift=True)
    # filter
    dist = dist / cutoff_freq
    return np.sqrt(1.0 / (1.0 + np.power(dist, fall)))


def gauss_bandpass(map_shape, sigma, center=0.0, keep_shape=False):
    """
    Bandpass filter with a gaussian function
    sigma : cutoff frequency [0:0.5]
    """
    # sigma = sigma/1.414
    # get frequency shells
    dist = make_fourier_shell(map_shape, keep_shape=keep_shape, fftshift=True)
    # filter
    return np.exp(-((dist - center) ** 2) / (2 * sigma * sigma))


def gauss_lowpass(map_shape, sigma, keep_shape=False):
    """
    Bandpass filter with a gaussian function
    sigma : cutoff frequency [0:0.5]
    """
    # sigma = sigma/1.414
    # get frequency shells
    dist = make_fourier_shell(map_shape, keep_shape=keep_shape, fftshift=True)
    # filter
    return np.exp(-(dist**2) / (2 * sigma * sigma))


def make_spherical_footprint(radius):
    """
    Get spherical footprint of a given diameter
    """
    rad_z = np.arange(radius * -1, radius + 1)
    rad_y = np.arange(radius * -1, radius + 1)
    rad_x = np.arange(radius * -1, radius + 1)

    rad_x = rad_x**2
    rad_y = rad_y**2
    rad_z = rad_z**2
    dist = np.sqrt(rad_z[:, None, None] + rad_y[:, None] + rad_x)
    # set_printoptions(threshold='nan')
    return (dist <= radius) * 1


def softmask_mean(arr, window=3, iter=5):
    footprint_sph = make_spherical_footprint(np.ceil(float(window) / 2))
    newarray = np.copy(arr)
    bin_arr = arr > 0.0
    added_edge_mask = bin_arr[:]
    bin_newarr = added_edge_mask
    added_edge = arr[bin_arr]
    for i in range(iter):
        newarray[:] = generic_filter(
            newarray, np.mean, footprint=footprint_sph, mode="constant", cval=0.0
        )
        newarray[added_edge_mask] = added_edge[:]
        added_edge_mask = np.logical_not(bin_newarr) * (newarray > 0.0)
        added_edge = newarray[added_edge_mask]
        bin_newarr = newarray > 0.0
    #         newarray[:] = uniform_filter(newarray,size=window,
    #                                      mode='constant',cval=0.0)
    newarray[bin_arr] = arr[bin_arr]
    return newarray


def softmask_gaussian(arr, sigma=1):
    newarray = np.copy(arr)
    bin_arr = arr > 0.0
    newarray[:] = gaussian_filter(newarray, sigma=sigma, mode="constant", cval=0.0)
    newarray[bin_arr] = arr[bin_arr]
    return newarray


def calculate_shell_correlation(shell1, shell2):
    cov_ps1_ps2 = shell1 * np.conjugate(shell2)
    sig_ps1 = shell1 * np.conjugate(shell1)
    sig_ps2 = shell2 * np.conjugate(shell2)
    cov_ps1_ps2 = np.sum(np.real(cov_ps1_ps2))
    var_ps1 = np.sum(np.real(sig_ps1))
    var_ps2 = np.sum(np.real(sig_ps2))
    # skip shells with no variance
    if np.round(var_ps1, 15) == 0.0 or np.round(var_ps2, 15) == 0.0:
        fsc = 0.0
    else:
        fsc = cov_ps1_ps2 / (np.sqrt(var_ps1 * var_ps2))
    return fsc


def calculate_fsc(
    ftarr1,
    ftarr2,
    dist1,
    dist2=None,
    step=None,
    maxlevel=None,
    map_apix=None,
    reso=None,
    plot=False,
):
    list_freq = []
    list_fsc = []
    list_nsf = []
    x = 0.0
    # for grids of different dimensions
    if dist2 is None:
        dist2 = dist1
    # Assume maxlevel is N/2 if maxlevel is None
    if step is None:
        assert compare_tuple(ftarr1.shape, ftarr2.shape)
        maxlevel = ftarr1.shape[0] // 2
    # Assume step=1 and dist is in range 0-N/2, if step is None
    if step is None:
        assert compare_tuple(ftarr1.shape, ftarr2.shape)
        step = 1.0
    highlevel = x + step

    while x < maxlevel:
        fshell_indices = (dist1 < min(maxlevel, highlevel)) & (dist1 >= x)
        fsc = calculate_shell_correlation(
            ftarr1[fshell_indices], ftarr2[fshell_indices]
        )
        # print highlevel, fsc
        list_freq.append(highlevel)
        list_fsc.append(fsc)
        num_nonzero_avg = min(
            np.count_nonzero(ftarr1[fshell_indices]),
            np.count_nonzero(ftarr2[fshell_indices]),
        )
        list_nsf.append(num_nonzero_avg)
        x = highlevel
        highlevel = x + step
    if step == 1.0 and map_apix is not None:
        list_freq = np.array(list_freq) / (ftarr1.shape[0] * map_apix)

    listfreq, listfsc, listnsf = zip(*sorted(zip(list_freq, list_fsc, list_nsf)))
    return listfreq, listfsc, listnsf


def scale_amplitudes(
    ftarr1, ftarr2, dist1, dist2=None, maxlevel=None, step=None, ref=False, plot=False
):
    """
    Scale amplitudes in each resolution shell:
        to avg (by default)
        to a reference (ref=True)
    dist1,dist2 : frequency shells (fftshifted) [0-0.5 or 0-N/2 or
                                                0-1/(nyquist resolution)]
    For grids of same dimensions, dist2 = dist1 (dist2=None)
    step: shell width (values in agreement with dist1,dist2)
            e.g. 1/N if dist in [0-0.5]
                1/(N*apix) if dist in [0-1/nyq.res]
                1 if dist in [0-N/2]
    maxlevel : nyquist (values in agreement with dist1,dist2)
    """
    # SCALING
    # storing for plots
    ft1_avg = []
    ft2_avg = []
    ft1_avg_new = []
    lfreq = []
    nc = 0
    x = 0.0
    highlevel = x + step
    # for grids of different dimensions
    if dist2 is None:
        dist2 = dist1
    # Assume step=1 and dist is in range 0-N/2, if step is None
    if step is None:
        assert compare_tuple(ftarr1.shape, ftarr2.shape)
        step = 1
    # Assume maxlevel is N/2 if maxlevel is None
    if maxlevel is None:
        assert compare_tuple(ftarr1.shape, ftarr2.shape)
        maxlevel = ftarr1.shape[0] // 2
    while x < maxlevel:
        # indices between upper and lower shell bound
        fshells1 = (dist1 < min(maxlevel, highlevel)) & (dist1 >= x)
        # radial average
        shellvec1 = ftarr1[fshells1]
        # indices between upper and lower shell bound
        fshells2 = (dist2 < min(maxlevel, highlevel)) & (dist2 >= x)
        # radial average
        shellvec2 = ftarr2[fshells2]

        abs1 = abs(shellvec1)
        abs2 = abs(shellvec2)
        # check non zero amplitudes in a shell
        ns1 = len(np.nonzero(abs1)[0])  # or count_nonzero
        ns2 = len(np.nonzero(abs2)[0])  # or count_nonzero
        # only few fourier terms in a shell: merge two shells
        if (ns1 < 5 or ns2 < 5) and nc < 3:
            nc += 1
            highlevel = min(maxlevel, x + (nc + 1) * step)
            if highlevel < maxlevel:
                continue
        else:
            nc = 0
        # scale intensities
        mft1 = np.mean(np.square(abs1))
        mft2 = np.mean(np.square(abs2))
        if mft1 == 0.0 and mft2 == 0.0:
            x = highlevel
            highlevel = x + step
            if highlevel < maxlevel:
                continue
        if plot:
            # sq of radial avg amplitude
            ft1_avg.append(np.log10(mft1))
            ft2_avg.append(np.log10(mft2))
        # scale to amplitudes of the ref map
        if ref:
            if mft1 == 0.0:
                continue
            ftarr1[fshells1] = shellvec1 * np.sqrt(mft2 / mft1)
        else:
            # replace with avg amplitudes for the two maps
            ftarr1[fshells1] = shellvec1 * np.sqrt((mft2 + mft1) / (2.0 * mft1))
            ftarr2[fshells2] = shellvec2 * np.sqrt((mft2 + mft1) / (2.0 * mft2))

        if plot:
            # new radial average (to check)
            mft1 = np.mean(
                np.square(abs(ftarr1[fshells1]))
            )  # numsum(absolute(ft1.fullMap[fshells1]))/len(shellvec1)
            ft1_avg_new.append(np.log10(mft1))
            lfreq.append(x + (highlevel - x) / 2.0)

        del fshells1, fshells2, shellvec1, shellvec2
        x = highlevel
        highlevel = x + step
        gc.collect()

    dict_plot = {}
    if plot:
        dict_plot["map1"] = [lfreq, ft1_avg]
        dict_plot["map2"] = [lfreq, ft2_avg]
        dict_plot["scaled"] = [lfreq, ft1_avg_new]
        return dict_plot


def get_median_deviation(array: np.ndarray) -> float:
    return np.median(np.absolute(array - np.median(array)))


def scale_median(arr1, arr2):
    """Scale one list/array of scores with respect to another based on
    distribution around median.

    Arguments:
        arr1, arr2: Array/list of scores

    Returns:
        Scaled arr1, based on the values in arr2
    """
    nparr1 = np.array(arr1)
    nparr2 = np.array(arr2)
    # median deviation
    med_dev_1 = get_median_deviation(nparr1)
    med_dev_2 = get_median_deviation(nparr2)
    if med_dev_1 == 0.0:
        scale_factor = 1.0
    else:
        scale_factor = med_dev_2 / med_dev_1
    shift_factor = np.median(nparr2) - (scale_factor * np.median(nparr1))

    # TODO: find a better way to avoid outliers in general
    if (max(nparr1) - min(nparr1)) > 0.1:
        scaled_arr = ((scale_factor * nparr1 + shift_factor) + nparr2) / 2.0
    else:
        scaled_arr = nparr1
    return scaled_arr


def calculate_array_overlap(
    array1,
    array2,
    arrays_masked=False,
    array1_threshold=None,
    array2_threshold=None,
):
    """
    Calculate fraction of overlap between two arrays (non-zero values)

    return:
    fraction of overlap with respect to the first array, with respect to
    second and with respect to the total (union)
    """
    if not arrays_masked and not (array1_threshold or array2_threshold):
        raise ValueError("Please provide masked maps or contour thresholds")
    elif arrays_masked:
        binmap1 = array1
        binmap2 = array2
    else:
        binmap1 = array1 > float(array1_threshold)
        binmap2 = array2 > float(array2_threshold)
    mask_array = (binmap1 * binmap2) > 0

    size1 = np.sum(binmap1)
    size2 = np.sum(binmap2)
    return (
        float(np.sum(mask_array)) / size1,
        float(np.sum(mask_array)) / size2,
        float(np.sum(mask_array)) / (size1 + size2),
    )


def compare_tuple(tuple1, tuple2):
    for val1, val2 in zip(tuple1, tuple2):
        if type(val2) is float:
            if round(val1, 2) != round(val2, 2):
                return False
        else:
            if val1 != val2:
                return False
    return True



#
#     Copyright (C) 2021 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.
#

import unittest
import os
import shutil
import tempfile
import numpy as np
from ccpem_utils_tests import test_data
from ccpem_pyutils.model.coord_grid import (
    set_map_grid,
    set_cubic_map_grid,
    mapGridPositions,
)
from ccpem_pyutils.map.parse_mrcmapobj import MapObjHandle
from ccpem_pyutils.scripts import shift_map_model_origin_zero
import mrcfile
import subprocess


class MapParseTests(unittest.TestCase):
    def setUp(self):
        """
        Setup test data and output directories.
        """
        self.test_data = os.path.dirname(test_data.__file__)
        self.test_dir = tempfile.mkdtemp(prefix="map_parse")
        # Change to test directory
        self._orig_dir = os.getcwd()
        os.chdir(self.test_dir)

    def tearDown(self):
        os.chdir(self._orig_dir)
        print(self.test_dir)
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_model_grid(self):
        # read
        model_input = os.path.join(self.test_data, "5me2.pdb")
        origin, dim = set_map_grid(model_input, apix=(1.0, 1.0, 1.0))
        with mrcfile.new("5me2.mrc") as mrc:
            mrc.header.origin.x = origin[0]
            mrc.header.origin.y = origin[1]
            mrc.header.origin.z = origin[2]
            # dimensions
            mrc.header.cella.x = dim[0]
            mrc.header.cella.y = dim[1]
            mrc.header.cella.z = dim[2]
            # voxel_size
            mrc.voxel_size = (1.0, 1.0, 1.0)
            mrc.set_data(np.zeros((dim[2], dim[1], dim[0]), dtype="float32"))
        assert dim == (108, 110, 96)
        origin, dim = set_cubic_map_grid(model_input, apix=(1.0, 1.0, 1.0))
        assert dim == (110, 110, 110)

    def test_run_subprocess_shift_map_model_zero(self):
        map_input = os.path.join(self.test_data, "emd_3488.mrc")
        with mrcfile.open(map_input, mode="r", permissive=True) as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
        # crop
        wrapped_mapobj.crop_map_grid(new_dim=(71, 73, 58))
        # write
        with mrcfile.new("emd_3488_cropped.mrc", overwrite=True) as mrc:
            wrapped_mapobj.update_newmap_data_header(mrc)
            list_grid_points = mapGridPositions(
                wrapped_mapobj, atom_coord=(61.197, 39.327, 61.266), res_map=3.2
            )[:]
        ox, oy, oz = wrapped_mapobj.origin
        wrapped_mapobj.close()
        model_input = os.path.join(self.test_data, "5me2.pdb")
        subprocess.call(
            [
                "python3 "
                + os.path.realpath(shift_map_model_origin_zero.__file__)
                + " -m "
                + "emd_3488_cropped.mrc"
                + " -p "
                + model_input,
            ],
            shell=True,
        )
        shifted_map = (
            os.path.splitext(os.path.basename("emd_3488_cropped.mrc"))[0]
            + "_shifted.mrc"
        )
        with mrcfile.open(shifted_map, mode="r", permissive=True) as mrc:
            wrapped_mapobj = MapObjHandle(mrc)
            list_grid_points_shifted = mapGridPositions(
                wrapped_mapobj,
                atom_coord=(61.197 - ox, 39.327 - oy, 61.266 - oz),
                res_map=3.2,
            )
        shifted_model = (
            os.path.splitext(os.path.basename(model_input))[0]
            + "_shifted"
            + os.path.splitext(model_input)[1]
        )
        assert os.path.isfile(shifted_map)
        assert os.path.isfile(shifted_model)
        assert np.array_equal(list_grid_points, list_grid_points_shifted)



import argparse
import gemmi
import numpy as np
import pandas as pd


def main(
    FOM=1.0,
    SIGF_scale=0.1,
    colin_fo="Fout",
    mtzin="nemap.mtz",
    mtzout="starting_nemap.mtz",
    spacegroup=None,
):
    # Set column FOM (W) to 1.0
    # Set column SIGF (Q) to 0.1 (=10%, default) of Fo values
    mtz = gemmi.read_mtz_file(mtzin)
    data_array = np.array(mtz, copy=True)
    data_frame = pd.DataFrame(data=data_array, columns=mtz.column_labels())
    if spacegroup is not None:
        mtz.spacegroup = gemmi.SpaceGroup(spacegroup)
    mtz.add_column("SIGF", "Q")
    data_frame["SIGF"] = SIGF_scale * data_frame[colin_fo]
    # data_frame['SIGF'] = 1.0
    mtz.add_column("FOM", "W")
    data_frame["FOM"] = FOM
    mtz.set_data(data_frame.to_numpy())
    mtz.history += ["Added columns SIGF and FOM using GEMMI."]
    mtz.write_to_file(mtzout)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-mtzin",
        "--mtzin",
        help=("Input MTZ filename"),
        type=str,
        dest="mtzin",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-mtzout",
        "--mtzout",
        help=("Output MTZ filename"),
        type=str,
        dest="mtzout",
        default="starting_nemap.mtz",
        required=True,
    )
    parser.add_argument(
        "-colin_fo",
        "--colin_fo",
        help=("Fobs column label name; Default = %(default)s"),
        type=str,
        dest="colin_fo",
        default="Fout",
        required=True,
    )
    parser.add_argument(
        "-FOM",
        "--FOM",
        help=("Figure of merit value; Default = %(default)s"),
        type=float,
        dest="FOM",
        default=1.0,
        required=False,
    )
    parser.add_argument(
        "-SIGF_scale",
        "--SIGF_scale",
        help=("Scale for SIGF values, Default = 0.1 (10%% of Fo values)."),
        type=float,
        dest="SIGF_scale",
        default=0.1,
        required=False,
    )
    parser.add_argument(
        "-spacegroup",
        "--spacegroup",
        help=("Set spacegroup, e.g. P1"),
        type=str,
        dest="spacegroup",
        default=None,
        required=False,
    )
    args = parser.parse_args()
    main(
        FOM=args.FOM,
        SIGF_scale=args.SIGF_scale,
        colin_fo=args.colin_fo,
        mtzin=args.mtzin,
        mtzout=args.mtzout,
        spacegroup=args.spacegroup,
    )



#
#     Copyright (C) 2019 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.

from ccpem_pyutils.map.mrc_map_utils import mrcmap_kdtree
import numpy as np
import mrcfile
from typing import Union
from ccpem_pyutils.map.parse_mrcmapobj import MapObjHandle
from ccpem_pyutils.model.gemmi_model_utils import GemmiModelUtils
import pathlib


def mapGridPositions(
    map_instance: Union[mrcfile.mrcfile.MrcFile, MapObjHandle],
    atom_coord: Union[list, tuple, np.ndarray],
    gauss: bool = True,
    res_map: float = 3.0,
    sim_sigma_coeff: float = 0.187,
    sigma_thr: float = 2.0,
    dist_search: float = 0.0,
):
    """
    Returns the indices of the nearest pixels to an atom as a list.

    Arguments:
        :map_instance:
            MrcFile or MapObjHandle Map instance.
        :atom_coord:
            3D Coordinate.
        :res_map:
            Map resolution
        :gauss:
            Whether to use resolution dependent gaussian blur to determine search space
        :sim_sigma_coeff:
        :sigma_thr:
            Threshold for gaussian sigma
        :dist_search:
            Additional distance to search when gauss is False
            Nearest voxel is returned when dist_search=0.0
    """
    if map_instance.__class__.__name__ == "MrcFile":
        origin = map_instance.header.origin.item()
        apix = map_instance.voxel_size.item()
    else:
        origin = map_instance.origin
        apix = map_instance.apix
    atom_x, atom_y, atom_z = atom_coord

    x_pos = int(round((atom_x - origin[0]) / apix[0], 0))
    y_pos = int(round((atom_y - origin[1]) / apix[1], 0))
    z_pos = int(round((atom_z - origin[2]) / apix[2], 0))
    if (
        (map_instance.x_size() >= x_pos >= 0)
        and (map_instance.y_size() >= y_pos >= 0)
        and (map_instance.z_size() >= z_pos >= 0)
    ):
        gridtree = mrcmap_kdtree(map_instance)[0]
        if gauss:
            # search all points withing 2.0sigma
            list_points = gridtree.query_radius(
                [
                    [
                        atom_x,
                        atom_y,
                        atom_z,
                    ]
                ],
                max(
                    [
                        min(apix),
                        sigma_thr * max(sim_sigma_coeff * res_map, 1.0),
                    ]
                ),
            )[0]
        elif dist_search != 0:
            # search withing the given radius
            list_points = gridtree.query_radius(
                [
                    [
                        atom_x,
                        atom_y,
                        atom_z,
                    ]
                ],
                max(
                    [
                        dist_search,
                        min(apix),
                    ]
                ),
            )[0]
        else:
            list_points = gridtree.query(
                [
                    [
                        atom_x,
                        atom_y,
                        atom_z,
                    ]
                ],
                k=1,
            )[1]
        return list_points
    else:
        print("Warning, atom out of map box")
        return []


def set_map_grid(
    modelfile: Union[str, pathlib.Path],
    apix: Union[tuple, list, np.ndarray],
    edge: Union[int, float] = 20.0,
):
    """
    Find grid parameters to fit the model

    Arguments:

    :modelfile: input model file
    :apix: voxel size of the output grid
    :edge: padding to the model extremities in Angstroms

    Return:
    Map grid origin and dimensions (number of voxels)
    """
    gemmiutils = GemmiModelUtils(modelfile)
    list_coords = gemmiutils.get_coordinates(return_list=True)[1]
    coords_array = np.array(list_coords)
    del list_coords
    list_x, list_y, list_z = np.transpose(coords_array)
    min_x = min(list_x) - edge
    max_x = max(list_x) + edge
    min_y = min(list_y) - edge
    max_y = max(list_y) + edge
    min_z = min(list_z) - edge
    max_z = max(list_z) + edge
    dim_x, dim_y, dim_z = max_x - min_x, max_y - min_y, max_z - min_z
    # adjust dimensions to number of voxels
    rem_x, rem_y, rem_z = dim_x % apix[0], dim_y % apix[1], dim_z % apix[2]
    dim_x, dim_y, dim_z = (
        dim_x + apix[0] - rem_x,
        dim_y + apix[1] - rem_y,
        dim_z + apix[2] - rem_z,
    )
    if int(dim_z / apix[2]) % 2 != 0:
        dim_z = dim_z + apix[2]  # z not even?
    if int(dim_x / apix[0]) % 2 != 0:
        dim_x = dim_x + apix[0]  # x not even?
    origin = min_x - apix[0] + rem_x, min_y - apix[1] + rem_y, min_z - apix[2] + rem_z
    return origin, (int(dim_x / apix[0]), int(dim_y / apix[1]), int(dim_z / apix[2]))


def set_cubic_map_grid(
    modelfile: Union[str, pathlib.Path],
    apix: Union[tuple, list, np.ndarray],
    edge: Union[int, float] = 20.0,
):
    """
    Find CUBIC grid parameters to fit the model

    Arguments:

    :modelfile: input model file
    :apix: voxel size of the output grid
    :edge: padding to the model extremities in Angstroms

    Return:
    Map grid origin and dimensions (number of voxels)
    """
    gemmiutils = GemmiModelUtils(modelfile)
    list_coords = gemmiutils.get_coordinates(return_list=True)[1]
    coords_array = np.array(list_coords)
    del list_coords
    list_x, list_y, list_z = np.transpose(coords_array)
    min_x = min(list_x) - edge
    max_x = max(list_x) + edge
    min_y = min(list_y) - edge
    max_y = max(list_y) + edge
    min_z = min(list_z) - edge
    max_z = max(list_z) + edge
    dim_x, dim_y, dim_z = max_x - min_x, max_y - min_y, max_z - min_z
    dim = [dim_x, dim_y, dim_z]
    max_dim = max(dim)
    # set cubic dimensions
    if max_dim - dim_x != 0.0:
        min_x = min_x - (max_dim - dim_x) / 2.0
        max_x = max_x + (max_dim - dim_x) / 2.0
    if max_dim - dim_y != 0.0:
        min_y = min_y - (max_dim - dim_y) / 2.0
        max_y = max_y + (max_dim - dim_y) / 2.0
    if max_dim - dim_z != 0.0:
        min_z = min_z - (max_dim - dim_z) / 2.0
        max_z = max_z + (max_dim - dim_z) / 2.0
    dim_x, dim_y, dim_z = max_x - min_x, max_y - min_y, max_z - min_z
    # adjust dimensions to number of voxels
    rem_x, rem_y, rem_z = dim_x % apix[0], dim_y % apix[1], dim_z % apix[2]
    dim_x, dim_y, dim_z = (
        dim_x + apix[0] - rem_x,
        dim_y + apix[1] - rem_y,
        dim_z + apix[2] - rem_z,
    )
    origin = min_x - apix[0] + rem_x, min_y - apix[1] + rem_y, min_z - apix[2] + rem_z
    return origin, (int(dim_x / apix[0]), int(dim_y / apix[1]), int(dim_z / apix[2]))



#
#     Copyright (C) 2021 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.
#

import unittest
import os
import math
import tempfile
import shutil
from ccpem_utils_tests import test_data
from ccpem_pyutils.model.gemmi_model_utils import (
    GemmiModelUtils,
    set_bfactor_attributes,
)
from ccpem_pyutils.scripts import analyse_bfactors
import subprocess


class GemmiModelUtilsTest(unittest.TestCase):
    def setUp(self):
        """
        Setup test data and output directories.
        """
        self.test_data = os.path.dirname(test_data.__file__)
        self.test_dir = tempfile.mkdtemp(prefix="coord_clust")
        # Change to test directory
        self._orig_dir = os.getcwd()
        os.chdir(self.test_dir)

    def tearDown(self):
        os.chdir(self._orig_dir)
        print(self.test_dir)
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_get_sequence(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        gemmiutils = GemmiModelUtils(model_input)
        dict_seq = gemmiutils.get_sequence_from_atom_records()
        assert len(dict_seq) == 4
        assert len(dict_seq["A"]) == 140
        dict_seq = gemmiutils.get_sequence_resnum_from_atom_records()[0]
        assert len(dict_seq) == 4
        assert len(dict_seq["A"]) == 140

    def test_get_bfact_deviation(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        gemmiutils = GemmiModelUtils(model_input)
        dict_bfact_dev = gemmiutils.get_avgbfact_deviation(save_obj=True)
        assert len(dict_bfact_dev["1"]) == 4
        assert math.isclose(dict_bfact_dev["1"]["A"]["201"][1], 8.31479, rel_tol=1e-3)
        for model in dict_bfact_dev:
            dict_attr = {}
            for chain in dict_bfact_dev[model]:
                for res_id in dict_bfact_dev[model][chain]:
                    residue_id = "_".join([model, chain, res_id])
                    dict_attr[residue_id] = dict_bfact_dev[model][chain][res_id][1]
            break
        set_bfactor_attributes(model_input, dict_attr)

    def test_run_subprocess_analyse_bfactors(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        subprocess.call(
            [
                "python3 "
                + os.path.realpath(analyse_bfactors.__file__)
                + " -p "
                + model_input
                + " -odir "
                + self.test_dir,
            ],
            shell=True,
        )
        assert os.path.isfile(os.path.join(self.test_dir, "5me2_residue_bfactors.json"))
        assert math.isclose(
            os.stat(os.path.join(self.test_dir, "5me2_residue_bfactors.json")).st_size,
            11661,
            rel_tol=0.05,
        )
        assert os.path.isfile(
            os.path.join(self.test_dir, "5me2_residue_coordinates.json")
        )
        assert math.isclose(
            os.stat(
                os.path.join(self.test_dir, "5me2_residue_coordinates.json")
            ).st_size,
            19383,
            rel_tol=0.05,
        )



import json
from json import JSONEncoder, JSONDecoder


class SimpleObjEncoder(JSONEncoder):
    """
    Classes to encode simple objects to json based on __dict__
    """

    def default(self, obj):
        try:
            return obj.__dict__
        except AttributeError:
            pass


def encode_simple_obj_to_json_dict(obj):
    """
    Encodes less complex classes as json
    """
    return json.dumps(obj, cls=SimpleObjEncoder)


def encode_simple_obj_to_json_file(obj, jsonfile):
    """
    Encodes less complex classes in a json file
    """
    print(obj)
    with open(jsonfile, "w") as oj:
        return json.dump(obj, oj, cls=SimpleObjEncoder)


def from_json(json_dict):
    """
    Placeholder for custom function to pass to the decoder
    """
    return json_dict


def decode_simple_obj_from_json(obj, func_from_json=from_json):
    """
    Encodes less complex classes as json
    """

    json_dict = json.loads(obj)
    return JSONDecoder(object_hook=func_from_json).decode(json_dict)



#
#     Copyright (C) 2019 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.

import argparse
from ccpem_pyutils.model import gemmi_model_utils
from ccpem_pyutils.map.parse_mrcmapobj import MapObjHandle
import mrcfile
import os


def parse_args():
    parser = argparse.ArgumentParser(description="CCP-EM model tools")
    parser.add_argument(
        "-m",
        "--map",
        required=True,
        help="Input map (MRC)",
    )
    parser.add_argument(
        "-p",
        "--model",
        required=False,
        help="Input atomic model file (PDB or mmCIF/PDBx)",
    )

    return parser.parse_args()


def main():
    args = parse_args()
    # Find map origin
    with mrcfile.open(args.map, mode="r") as mrc:
        wrapped_mapobj = MapObjHandle(mrc)
        wrapped_mapobj.fix_origin()  # if non-zero nstart and zero origin
        ox, oy, oz = wrapped_mapobj.origin
        wrapped_mapobj.origin = (0.0, 0.0, 0.0)
        wrapped_mapobj.nstart = (0, 0, 0)
    shifted_map = os.path.splitext(os.path.basename(args.map))[0] + "_shifted.mrc"
    with mrcfile.new(shifted_map, overwrite=True) as mrc:
        wrapped_mapobj.update_newmap_data_header(mrc)
    wrapped_mapobj.close()
    shifted_model = (
        os.path.splitext(os.path.basename(args.model))[0]
        + "_shifted"
        + os.path.splitext(args.model)[1]
    )
    gemmiutils = gemmi_model_utils.GemmiModelUtils(args.model)
    gemmiutils.shift_coordinates(
        trans_vector=(-ox, -oy, -oz), out_model_path=shifted_model, remove_charges=False
    )


if __name__ == "__main__":
    main()



from __future__ import annotations
import numpy as np
import copy
from typing import Union, Sequence
from ccpem_pyutils.map.array_utils import crop_array, pad_array


class MapObjHandle(object):
    def __init__(self, mapobject, datacopy=True):
        if type(mapobject) is MapObjHandle:
            super(MapObjHandle, self).__init__()
            if not datacopy:
                self.__dict__ = mapobject.__dict__.copy()
            else:
                try:
                    mapobject.close()
                    self.mrc = mapobject.mrc
                except AttributeError:  # input map object detached
                    pass
                for k, v in mapobject.__dict__.items():
                    try:
                        self.__dict__[k] = copy.deepcopy(v)
                    except TypeError:  # skip file object pickle failures
                        pass
        else:
            super(MapObjHandle, self).__init__()
            # map data and header details
            # add other attributes?
            self.mrc = mapobject
            self.class_name = self.mrc.__class__.__name__
            if datacopy:
                self.initialize(inplace=False)
            else:
                self.initialize(inplace=True)

    def shape(self):
        return self.data.shape

    def min(self):
        return np.amin(self.data)

    def max(self):
        return np.amax(self.data)

    def std(self):
        return np.std(self.data)

    def x_size(self):
        """
        size of the map array in x direction.
        """
        return self.data.shape[2]

    def y_size(self):
        """
        size of the map array in y direction.
        """
        return self.data.shape[1]

    def z_size(self):
        """
        size of the map array in z direction.
        """
        return self.data.shape[0]

    def copy(self, deep=True, detach=True):
        """
        Copy contents to a new object
        """
        # create MapEdit object
        copymap = MapObjHandle(self)
        # detach from initial mrc mapfile
        if detach:
            copymap.mrc = None
        # copy data and header
        copymap.origin = copy.deepcopy(self.origin)
        if deep:
            copymap.data = self.data.copy()
        else:
            copymap.data = self.data
        copymap.apix = copy.deepcopy(self.apix)
        copymap.dim = copy.deepcopy(self.dim)
        return copymap

    def data_copy(self):
        """
        Copy data (not to modify inplace)
        """
        if self.class_name == "MrcFile":
            self.data = self.mrc.data.copy()
        elif self.class_name == "Map":
            self.data = self.mrc.fullMap.copy()

    def reinitialize_data(self, inplace=False):
        """
        Initialize or re-initialize data
        """
        if inplace:
            if self.class_name == "MrcFile":
                self.data = self.mrc.data
            elif self.class_name == "Map":
                self.data = self.mrc.fullMap
        else:
            self.data_copy()

    def reinitialize_header(self):
        """
        Initialize or re-initialize header
        """
        if self.class_name == "MrcFile":
            self.origin = self.mrc.header.origin.item()
            self.apix = self.mrc.voxel_size.item()
            self.dim = self.mrc.header.cella.item()
            self.nstart = (
                self.mrc.header.nxstart,
                self.mrc.header.nystart,
                self.mrc.header.nzstart,
            )
        elif self.class_name == "Map":
            self.origin = tuple(self.mrc.origin)
            self.apix = (
                round(self.mrc.header[10] / self.mrc.header[7], 2),
                round(self.mrc.header[11] / self.mrc.header[8], 2),
                round(self.mrc.header[12] / self.mrc.header[9], 2),
            )
            self.dim = (
                self.x_size() * self.apix[0],
                self.y_size() * self.apix[1],
                self.z_size() * self.apix[2],
            )
            self.nstart = self.mrc.header[4:7]
        else:
            raise TypeError(
                """Only MrcFile and TEMPY Map objects currently
                    supported"""
            )

    def initialize(self, inplace=False):
        """
        Initialize/re-initialize data/header
        """
        self.reinitialize_data(inplace=inplace)
        self.reinitialize_header()

    # update map header records
    def update_input_header(self):
        """
        Update map header records to current values
        """
        if self.class_name == "MrcFile":
            # origin
            self.mrc.header.origin.x = self.origin[0]
            self.mrc.header.origin.y = self.origin[1]
            self.mrc.header.origin.z = self.origin[2]
            # dimensions
            self.mrc.header.cella.x = self.dim[0]
            self.mrc.header.cella.y = self.dim[1]
            self.mrc.header.cella.z = self.dim[2]
            # voxel_size
            if type(self.apix) is float:
                self.mrc.voxel_size = (self.apix, self.apix, self.apix)
            else:
                self.mrc.voxel_size = self.apix
            self.mrc.header.nxstart = self.nstart[0]
            self.mrc.header.nystart = self.nstart[1]
            self.mrc.header.nzstart = self.nstart[2]
        elif self.class_name == "Map":
            # origin
            self.mrc.origin[0] = self.origin[0]
            self.mrc.origin[1] = self.origin[1]
            self.mrc.origin[2] = self.origin[2]
            # tempy takes a single voxel_size [TODO]
            self.mrc.apix = self.apix[0]
            self.mrc.header[4] = self.nstart[0]
            self.mrc.header[5] = self.nstart[1]
            self.mrc.header[6] = self.nstart[2]

    @staticmethod
    def compare_tuple(tuple1, tuple2):
        for val1, val2 in zip(tuple1, tuple2):
            if type(val2) is float:
                if round(val1, 2) != round(val2, 2):
                    return False
            else:
                if val1 != val2:
                    return False
        return True

    # update map array data
    def update_input_data(self):
        """
        Update input data array with current
        """
        if self.class_name == "MrcFile":
            if self.compare_tuple(self.data.shape, self.mrc.data.shape):
                self.mrc.data[:] = self.data
            else:
                self.mrc.set_data(self.data)
        elif self.class_name == "Map":
            self.mrc.fullMap[:] = self.data

    def update_input_data_header(self):
        """
        Update data and header with current values
        """
        self.update_input_data()
        self.update_input_header()

    def close(self):
        if self.class_name == "MrcFile":
            self.mrc.close()
        elif self.class_name == "Map":
            self.mrc.fullMap = None
            self.mrc = None

    # update map header records
    def update_newmap_header(self, newmap):
        """
        Copy current values to a newmap (mrcfile/ TEMPy Map obj)
        """
        if newmap.__class__.__name__ == "MrcFile":
            # origin
            newmap.header.origin.x = self.origin[0]
            newmap.header.origin.y = self.origin[1]
            newmap.header.origin.z = self.origin[2]
            # dimensions
            newmap.header.cella.x = self.dim[0]
            newmap.header.cella.y = self.dim[1]
            newmap.header.cella.z = self.dim[2]
            # voxel_size
            newmap.voxel_size = self.apix
            newmap.header.nxstart = self.nstart[0]
            newmap.header.nystart = self.nstart[1]
            newmap.header.nzstart = self.nstart[2]

        elif newmap.__class__.__name__ == "Map":
            # origin
            newmap.origin[0] = self.origin[0]
            newmap.origin[1] = self.origin[1]
            newmap.origin[2] = self.origin[2]
            # voxel_size
            newmap.apix = self.apix[0]
            newmap.header[4] = self.nstart[0]
            newmap.header[5] = self.nstart[1]
            newmap.header[6] = self.nstart[2]

    # update map array data
    def update_newmap_data(self, newmap):
        """
        Update new map (mrcfile/TEMPy) data array with current
        """
        if newmap.__class__.__name__ == "MrcFile":
            if str(self.data.dtype) == "float64":
                newmap.set_data(self.data.astype("float32", copy=False))
            else:
                newmap.set_data(self.data)
        elif newmap.__class__.__name__ == "Map":
            newmap.fullMap[:] = self.data

    def update_newmap_data_header(self, newmap):
        """
        Update data and header of mrcfile map obj with current values
        """
        self.update_newmap_data(newmap)
        self.update_newmap_header(newmap)

    def update_header_by_data(self):
        self.dim = (
            self.x_size() * self.apix[0],
            self.y_size() * self.apix[1],
            self.z_size() * self.apix[2],
        )

    def set_attributes_tempy(self):
        """
        Set class attributes to use with TEMPy functions
        """
        self.fullMap = self.data
        self.nxstart = self.nstart[0]
        self.nystart = self.nstart[1]
        self.nzstart = self.nstart[2]

    def set_dim_apix(self, apix):
        """
        Set dimensions (Angstroms) given voxel size
        """
        self.apix = apix
        self.dim = (
            self.x_size() * self.apix[0],
            self.y_size() * self.apix[1],
            self.z_size() * self.apix[2],
        )

    def set_apix_dim(self, dim):
        """
        Set voxel size given dimensions (Angstroms) of Grid
        """
        self.dim = dim
        self.apix = (
            np.around(self.dim[0] / self.x_size(), decimals=3),
            np.around(self.dim[1] / self.y_size(), decimals=3),
            np.around(self.dim[2] / self.z_size(), decimals=3),
        )

    def set_apix_tempy(self):
        """
        Set apix to single float value for using TEMPy functions
        """
        if isinstance(self.apix, tuple):
            if self.apix[0] == self.apix[1] == self.apix[2]:
                self.apix = self.apix[0]
            else:
                self.downsample_apix(max(self.apix), inplace=True)
                self.apix = self.apix[0]

    def shift_origin(self, new_origin: Union[list, tuple, np.ndarray]):
        """
        Shift map to given origin
        """
        assert len(new_origin) == 3
        self.origin = tuple(new_origin)

    def shift_nstart(self, new_nstart: Union[list, tuple, np.ndarray]):
        """
        Update nstart record to given nstart
        """
        assert len(new_nstart) == 3
        self.nstart = tuple(new_nstart)

    def fix_origin(self):
        """
        Set origin record based on nstart if non-zero
        """
        if self.origin[0] == 0.0 and self.origin[1] == 0.0 and self.origin[2] == 0.0:
            if self.nstart[0] != 0 or self.nstart[1] != 0 or self.nstart[2] != 0:
                self.set_apix_as_tuple()
                # origin
                self.origin = (
                    self.nstart[0] * self.apix[0],
                    self.nstart[1] * self.apix[1],
                    self.nstart[2] * self.apix[2],
                )

    def crop_map_grid(
        self,
        new_dim: Sequence[int] = None,
        contour: float = None,
        ext: int = None,
        cubic: bool = False,
        inplace: bool = True,
        mask_map: MapObjHandle = None,
    ):
        """
        Crop grid based on given dimensions or contour or mask

        Arguments
        ---------
        :arr: np.ndarray
            input array (indexed as arr[z][y][x])
        :new_dim:
            new dimensions to crop to
        :contour: float, optional
            map threshold (slices with all values lower than this are cropped)
        :ext: int, optional
            number of pixels/voxels as padding
        :mask_array: np.ndarray
            mask to be applied before cropping (slices with all zeros are cropped)
        :cubic: bool
            whether to crop to a cubic box?
        :inplace: bool
            whether to modify existing map object inplace or create new?

        Return
        ------
        cropped map object if not inplace
        """
        if mask_map:
            cropped_array, start = crop_array(
                arr=self.data,
                new_dim=new_dim,
                contour=contour,
                cubic=cubic,
                ext=ext,
                mask_array=mask_map.data,
            )
        else:
            cropped_array, start = crop_array(
                arr=self.data,
                new_dim=new_dim,
                contour=contour,
                cubic=cubic,
                ext=ext,
            )
        ox = self.origin[0] + start[0] * self.apix[0]
        oy = self.origin[1] + start[1] * self.apix[1]
        oz = self.origin[2] + start[2] * self.apix[2]
        nstx = self.nstart[0] + start[0]
        nsty = self.nstart[1] + start[1]
        nstz = self.nstart[2] + start[2]
        if inplace:
            self.origin = (ox, oy, oz)
            self.data = cropped_array
            self.nstart = (nstx, nsty, nstz)
            self.update_header_by_data()
        else:
            newmap = self.copy()
            newmap.origin = (ox, oy, oz)
            newmap.data = cropped_array
            newmap.nstart = (nstx, nsty, nstz)
            newmap.update_header_by_data()
            return newmap

    def pad_map_grid(
        self,
        ext_dim: Sequence[int] = None,
        inplace: bool = True,
    ):
        """
        Pad grid based on given dimensions

        Arguments
        ---------
        :arr: np.ndarray
            input array (indexed as arr[z][y][x])
        :ext_dim:
            extension to add to each dimension (nx,ny,nz)
        :inplace: bool
            whether to modify existing map object inplace or create new?

        Return
        ------
        Padded map object if not inplace
        """
        padded_array, start = pad_array(
            arr=self.data, nx=ext_dim[0], ny=ext_dim[1], nz=ext_dim[2]
        )
        # shift origin
        ox = self.origin[0] - start[0] * self.apix[0]
        oy = self.origin[1] - start[1] * self.apix[1]
        oz = self.origin[2] - start[2] * self.apix[2]
        nstx = self.nstart[0] - start[0]
        nsty = self.nstart[1] - start[1]
        nstz = self.nstart[2] - start[2]
        if inplace:
            self.origin = (ox, oy, oz)
            self.data = padded_array
            self.nstart = (nstx, nsty, nstz)
            self.update_header_by_data()
        else:
            newmap = self.copy()
            newmap.origin = (ox, oy, oz)
            newmap.data = padded_array
            newmap.nstart = (nstx, nsty, nstz)
            newmap.update_header_by_data()
            return newmap

    def set_apix_as_tuple(self):
        if isinstance(self.apix, (int, float)):
            self.apix = (self.apix, self.apix, self.apix)



#
#     Copyright (C) 2021 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.
#

import unittest
import os
import tempfile
import shutil
from ccpem_utils_tests import test_data
import numpy as np
from ccpem_pyutils.model.gemmi_model_utils import (
    GemmiModelUtils,
    get_residue_attribute,
    set_bfactor_attributes,
)
from ccpem_pyutils.other.cluster import (
    cluster_coord_features,
    generate_kdtree,
    pairs_kdtree,
)
import datetime


class CoordClustTests(unittest.TestCase):
    def setUp(self):
        """
        Setup test data and output directories.
        """
        self.test_data = os.path.dirname(test_data.__file__)
        self.test_dir = tempfile.mkdtemp(prefix="coord_clust")
        # Change to test directory
        self._orig_dir = os.getcwd()
        os.chdir(self.test_dir)

    def tearDown(self):
        os.chdir(self._orig_dir)
        print(self.test_dir)
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def gemmi_get_coord(self, return_list=True, atom_selection="all"):
        list_ids, list_coordinates = self.gemmiutils.get_coordinates(
            return_list=return_list, atom_selection=atom_selection
        )
        return list_ids, list_coordinates

    def cluster_coord_features(self, list_coordinates, dbscan_eps=2.3):
        cluster_labels = cluster_coord_features(
            np.array(list_coordinates), dbscan_eps=dbscan_eps, norm=False
        )
        return cluster_labels

    def cluster_coord_features_norm(self, list_coordinates, dbscan_eps=0.035):
        cluster_labels = cluster_coord_features(
            np.array(list_coordinates), dbscan_eps=dbscan_eps, norm=True
        )
        return cluster_labels

    def test_cluster_model_coordinates(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        self.gemmiutils = GemmiModelUtils(model_input)
        list_coordinates = self.gemmi_get_coord()[1]
        assert len(list_coordinates) == 4468
        cluster_labels = self.cluster_coord_features(list_coordinates)
        assert len(cluster_labels) == len(list_coordinates)
        assert len(np.unique(cluster_labels)) == 8

    def test_cluster_model_coordinates_norm(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        self.gemmiutils = GemmiModelUtils(model_input)
        list_coordinates = self.gemmi_get_coord()[1]
        assert len(list_coordinates) == 4468
        cluster_labels = self.cluster_coord_features_norm(list_coordinates)
        assert len(cluster_labels) == len(list_coordinates)
        assert len(np.unique(cluster_labels)) == 8

    def test_cluster_model_coordinates_backbone(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        self.gemmiutils = GemmiModelUtils(model_input)
        list_ids, list_coordinates = self.gemmi_get_coord(atom_selection="backbone")
        assert len(list_coordinates) == 2436
        cluster_labels = self.cluster_coord_features(
            list_coordinates, dbscan_eps=2.8
        )  # norm: 0.04
        assert len(cluster_labels) == len(list_coordinates)
        assert len(np.unique(cluster_labels)) == 8

    def test_cluster_model_coordinates_representative(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        self.gemmiutils = GemmiModelUtils(model_input)
        list_ids, list_coordinates = self.gemmi_get_coord(
            atom_selection="one_per_residue"
        )
        assert len(list_coordinates) == 570
        cluster_labels = self.cluster_coord_features(
            list_coordinates, dbscan_eps=3.83
        )  # norm 0.07673
        assert len(cluster_labels) == len(list_coordinates)
        assert len(np.unique(cluster_labels)) == 19
        list_res_ids, list_res_attr = get_residue_attribute(list_ids, cluster_labels)
        dict_attr = {}
        for n in range(len(list_res_attr)):
            dict_attr[list_res_ids[n]] = list_res_attr[n]
        set_bfactor_attributes(model_input, dict_attr)

    def test_cluster_model_coordinates_centre(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        self.gemmiutils = GemmiModelUtils(model_input)
        list_ids, list_coordinates = self.gemmi_get_coord(atom_selection="centre")
        assert len(list_coordinates) == 570
        cluster_labels = self.cluster_coord_features(
            list_coordinates, dbscan_eps=5.35
        )  # norm 0.088
        assert len(cluster_labels) == len(list_coordinates)
        assert len(np.unique(cluster_labels)) == 30

    def test_kdtree_model_coordinates_pairs(self):
        model_input = os.path.join(self.test_data, "5me2.pdb")
        self.gemmiutils = GemmiModelUtils(model_input)
        list_coordinates = self.gemmi_get_coord(atom_selection="one_per_residue")[1]
        assert len(list_coordinates) == 570
        print("Start: ", datetime.datetime.now())
        kdtree = generate_kdtree(list_coordinates, leaf_size=25)
        pair_indices = pairs_kdtree([(61.197, 39.327, 61.266)], kdtree, dist=5.0)
        print("End: ", datetime.datetime.now())
        assert len(pair_indices) == 6



#
#     Copyright (C) 2019 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.

from ccpem_pyutils.other.cluster import generate_kdtree
import numpy as np


def mrcmap_kdtree(map_instance):
    """
    Returns the KDTree of coordinates from a mrc map grid.

    Arguments:
        *map_instance*
            MrcFile or MapObjHandle Map instance.
    """
    if map_instance.__class__.__name__ == "MrcFile":
        origin = map_instance.header.origin.item()
        apix = map_instance.voxel_size.item()
        nz, ny, nx = map_instance.data.shape
    else:
        origin = map_instance.origin
        apix = map_instance.apix
        nz, ny, nx = map_instance.data.shape

    # convert to real coordinates
    zg, yg, xg = np.mgrid[0:nz, 0:ny, 0:nx]
    # to get indices in real coordinates
    zg = zg * apix[2] + origin[2] + apix[2] / 2.0
    yg = yg * apix[1] + origin[1] + apix[1] / 2.0
    xg = xg * apix[0] + origin[0] + apix[0] / 2.0
    indi = list(zip(xg.ravel(), yg.ravel(), zg.ravel()))
    gridtree = generate_kdtree(indi, leaf_size=42)
    return gridtree, indi



#
#     Copyright (C) 2019 CCP-EM
#
#     This code is distributed under the terms and conditions of the
#     CCP-EM Program Suite Licence Agreement as a CCP-EM Application.
#     A copy of the CCP-EM licence can be obtained by writing to the
#     CCP-EM Secretary, RAL Laboratory, Harwell, OX11 0FA, UK.

import argparse
from ccpem_pyutils.model import gemmi_model_utils


def parse_args():
    parser = argparse.ArgumentParser(description="CCP-EM model tools")
    parser.add_argument(
        "-m",
        "--model",
        required=True,
        help="Input atomic model file (PDB or mmCIF/PDBx",
    )

    return parser.parse_args()


def main():
    args = parse_args()

    # Read input
    print("Reading input model file:", args.model)
    gemmi_model_utils.get_residue_ca_coordinates(
        in_model_path=args.model, prot_resatom="CA", dist_pairs=7.0
    )


if __name__ == "__main__":
    main()
